#! /usr/bin/env python

import os, tempfile, subprocess, shutil
import pysam
import parasail

from .pyssw import *
from .my_seq import reverse_complement
# from onebreak.long_read_validate import ssw_check
# from onebreak.my_seq import reverse_complement

def ssw_check(query, target, use_ssw_lib):

    if use_ssw_lib:
        return(ssw_check_ssw_lib(query, target))
    else:
        return(ssw_check_parasail(query, target))


def ssw_check_ssw_lib(target, query):

    nMatch = 2
    nMismatch = 2
    nOpen = 3
    nExt = 1
    sLibPath = ""
 
    lEle = []
    dRc = {} 
    dEle2Int = {}
    dInt2Ele = {}
    # init DNA score matrix
    lEle = ['A', 'C', 'G', 'T', 'N']
    dRc = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N': 'N', 'a':'T', 'c':'G', 'g':'C', 't':'A', 'n': 'N'} 
    for i,ele in enumerate(lEle):
        dEle2Int[ele] = i
        dEle2Int[ele.lower()] = i
        dInt2Ele[i] = ele
    nEleNum = len(lEle)
    lScore = [0 for i in range(nEleNum**2)]
    for i in range(nEleNum-1):
        for j in range(nEleNum-1):
            if lEle[i] == lEle[j]:
                lScore[i*nEleNum+j] = nMatch
            else:
                lScore[i*nEleNum+j] = -nMismatch

    # translate score matrix to ctypes
    mat = (len(lScore) * ct.c_int8) ()
    mat[:] = lScore

    # set flag
    nFlag = 1

    # check whether libssw.so is in LD_LIBRARY_PATH
    sLibPath = ""
    for ld_path in os.environ["LD_LIBRARY_PATH"].split(':'):
        # print ld_path
        if os.path.exists(ld_path + "/libssw.so"):
            sLibPath = ld_path # + "/libssw.so"
            break
    if sLibPath == "":
        print("cannot find libssw.so in LD_LIBRARY_PATH", file = sys.stderr)
        sys.exit(1)

    ssw = ssw_lib.CSsw(sLibPath)
    # supporting_reads = []
    alignment_info = {}

    # iterate query sequence
    for sQId,sQSeq,sQQual in read(query):

        sQSeq = sQSeq.strip('\n')

        # build query profile
        qNum = to_int(sQSeq, lEle, dEle2Int)
        qProfile = ssw.ssw_init(qNum, ct.c_int32(len(sQSeq)), mat, len(lEle), 2)

        # for reverse complement
        sQRcSeq = ''.join([dRc[x] for x in sQSeq[::-1]])
        qRcNum = to_int(sQRcSeq, lEle, dEle2Int)
        qRcProfile = ssw.ssw_init(qRcNum, ct.c_int32(len(sQSeq)), mat, len(lEle), 2)

        # set mask len
        if len(sQSeq) > 30:
            nMaskLen = int(len(sQSeq) / 2)
        else:
            nMaskLen = 15

        # iter target sequence
        for sRId,sRSeq,_ in read(target):

            rNum = to_int(sRSeq, lEle, dEle2Int)

            # format ofres: (nScore, nScore2, nRefBeg, nRefEnd, nQryBeg, nQryEnd, nRefEnd2, nCigarLen, lCigar)
            res = align_one(ssw, qProfile, rNum, len(sRSeq), nOpen, nExt, nFlag, nMaskLen)

            # align rc query
            resRc = align_one(ssw, qRcProfile, rNum, len(sRSeq), nOpen, nExt, nFlag, nMaskLen)
    
            # build cigar and trace back path
            if resRc == None or res[0] > resRc[0]:
                resPrint = res
                rstart, rend = resPrint[2] + 1, resPrint[3] + 1
                qstart, qend = resPrint[4] + 1, resPrint[5] + 1
                strand = '+'
                sCigar, sQ, sA, sR = buildPath(sQSeq, sRSeq, res[4], res[2], res[8])
            else:
                resPrint = resRc
                rstart, rend = resPrint[2] + 1, resPrint[3] + 1
                qstart, qend = len(sQSeq) - resPrint[5], len(sQSeq) - resPrint[4] 
                strand = '-'
                sCigar, sQ, sA, sR = buildPath(sQRcSeq, sRSeq, resRc[4], resRc[2], resRc[8])

            # if int(resPrint[0]) > score_ratio_thres * len(sRSeq) and int(resPrint[2]) + 1 < start_pos_thres * len(sRSeq) and int(resPrint[3]) + 1 > end_pos_thres * len(sRSeq):
            # supporting_reads.append([sQId, resPrint[0], resPrint[2] + 1, resPrint[3] + 1])
            # alignment_info[sQId] = [resPrint[0], resPrint[2] + 1, resPrint[3] + 1, resPrint[4] + 1, resPrint[5] + 1, strand]
            alignment_info[sQId] = [resPrint[0], rstart, rend, qstart, qend, strand]
        ssw.init_destroy(qProfile)
        ssw.init_destroy(qRcProfile)
       
    # return(supporting_reads)
    return(alignment_info)


def ssw_check_parasail(query, target):

    user_matrix = parasail.matrix_create("ACGT", 2, -2)

    alignment_info = {}
    for sQId, sQSeq, sQQual in read(query):
    
        sQSeq_r = reverse_complement(sQSeq)

        for sTId, sTSeq, STQual in read(target):

            res = parasail.ssw(sQSeq, sTSeq, 3, 1, user_matrix)
            res_r = parasail.ssw(sQSeq_r, sTSeq, 3, 1, user_matrix)

            if res.score1 > res_r.score1:
                score = res.score1
                qstart, qend = res.read_begin1 + 1., res.read_end1 + 1
                tstart, tend = res.ref_begin1 + 1, res.ref_end1 + 1
                strand = '+'
            else:
                score = res_r.score1
                qstart, qend = len(sQSeq) - res_r.read_end1, len(sQSeq) - res_r.read_begin1
                tstart, tend = res_r.ref_begin1 + 1, res_r.ref_end1 + 1
                strand = '-'

            alignment_info[sTId] = [score, int(qstart), int(qend), int(tstart), int(tend), strand]

    return(alignment_info)


def long_read_validate_by_alignment(sv_file, output_file, bam_file, reference, use_ssw_lib, debug, variant_sread_min_mapq = 0, validate_sequence_length = 200, score_ratio_thres = 1.4, start_pos_thres = 0.2, end_pos_thres = 0.8, var_ref_margin_thres = 10):
    
    def is_short_del_dup(key):
        keys = key.split(',')
        if keys[6] == "---": keys[6] == ''
        if keys[0] == keys[3] and keys[2] == '+' and keys[5] == '-' and int(keys[4]) - int(keys[1]) + len(keys[6]) < 100:
            return(True)
        elif keys[0] == keys[3] and keys[2] == '-' and keys[5] == '+' and int(keys[4]) - int(keys[1]) + len(keys[6]) < 100:
            return(True)
        else:
            return(False)

    
    bam_ps = pysam.AlignmentFile(bam_file, "rb")

    rname2key = {}
    key2rname2mapq = {}
    with open(sv_file, 'r') as hin:
        for line in hin:
            if line.startswith("#") or line.startswith("Chr_1"): continue 

            F = line.rstrip('\n').split('\t')               
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6]
            if tinseq == "---": tinseq = ''
            key = ','.join([tchr1, str(tpos1), tdir1, tchr2, str(tpos2), tdir2, str(len(tinseq))])

            if key not in key2rname2mapq: key2rname2mapq[key] = {}
            
            for read in bam_ps.fetch(tchr1, max(tpos1 - 100, 0), tpos1 + 100):

                if read.qname not in rname2key: rname2key[read.qname] = []
                rname2key[read.qname].append(key)

                if read.qname not in key2rname2mapq[key]: key2rname2mapq[key][read.qname] = [None, None]
                key2rname2mapq[key][read.qname][0] = read.mapping_quality 

            for read in bam_ps.fetch(tchr2, max(tpos2 - 100, 0), tpos2 + 100):

                if read.qname not in rname2key: rname2key[read.qname] = []
                rname2key[read.qname].append(key)

                if read.qname not in key2rname2mapq[key]: key2rname2mapq[key][read.qname] = [None, None]
                key2rname2mapq[key][read.qname][1] = read.mapping_quality



    # remove duplicated keys
    for rname in rname2key:
        keys = list(set(rname2key[rname]))
        rname2key[rname] = keys

    hout = open(output_file + ".tmp3.long_read_seq.unsorted", 'w')
    for read in bam_ps.fetch():

        flags = format(int(read.flag), "#014b")[:1:-1]
    
        # skip supplementary alignment
        if flags[8] == "1" or flags[11] == "1": continue

        # skip duplicated reads             
        if flags[10] == "1": continue

        if read.qname in rname2key:

            read_seq = read.query_sequence
            if flags[4] == "1": read_seq = reverse_complement(read_seq)

            for key in rname2key[read.qname]:

                print(key + '\t' + read.qname + '\t' + read_seq, file = hout)

    hout.close()
    bam_ps.close()

    hout = open(output_file + ".tmp3.long_read_seq.sorted", 'w')
    subprocess.call(["sort", "-k1,1", output_file + ".tmp3.long_read_seq.unsorted"], stdout = hout)
    hout.close()
    subprocess.call(["rm" ,"-rf", output_file + ".tmp3.long_read_seq.unsorted"])

    # my_seq.get_seq function could be used. But this procedure is repeatead many times and using pysam class may be good for the IO.
    reference_fasta = pysam.FastaFile(os.path.abspath(reference))
 
    key2contig = {}
    with open(sv_file, 'r') as hin:

        for line in hin:
            if line.startswith("#") or line.startswith("Chr_1"): continue

            F = line.rstrip('\n').split('\t')

            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6]
            if tinseq == "---": tinseq = ''
            key = ','.join([tchr1, str(tpos1), tdir1, tchr2, str(tpos2), tdir2, str(len(tinseq))])
 
            # reference_local_seq
            reference_local_seq_1 = reference_fasta.fetch(tchr1, max(tpos1 - validate_sequence_length - 1, 0), tpos1 + validate_sequence_length - 1) 
            reference_local_seq_2 = reference_fasta.fetch(tchr2, max(tpos2 - validate_sequence_length - 1, 0), tpos2 + validate_sequence_length - 1)  

            # variant_seq
            variant_seq = ""
            if tdir1 == '+':
                tseq = reference_fasta.fetch(tchr1, max(tpos1 - validate_sequence_length - 1, 0), tpos1 - 1)
            else:  
                tseq = reference_fasta.fetch(tchr1, tpos1 - 1, tpos1 + validate_sequence_length - 1)
                tseq = reverse_complement(tseq)

            if tdir1 == "+":
                variant_seq = tseq + tinseq
            else:
                variant_seq = tseq + reverse_complement(tinseq)

            if tdir2 == '-':
                tseq = reference_fasta.fetch(tchr2, tpos2 - 1, tpos2 + validate_sequence_length - 1)
            else:
                tseq = reference_fasta.fetch(tchr2, max(tpos2 - validate_sequence_length - 1, 0), tpos2 - 1)
                tseq = reverse_complement(tseq)

            variant_seq = variant_seq + tseq

            variant_seq_1 = variant_seq[:min(2 * validate_sequence_length, len(variant_seq))]
            variant_seq_2 = variant_seq[-min(2 * validate_sequence_length, len(variant_seq)):]

            # print(key)
            # key2contig[key] = [variant_seq, reference_local_seq_1,reference_local_seq_2]
            # if add_reference_alignment:
            #     key2contig[key] = [variant_seq_1, variant_seq_2, reference_local_seq_1,reference_local_seq_2]
            # else:
            #     key2contig[key] = [variant_seq_1, variant_seq_2]
            key2contig[key] = [variant_seq_1, variant_seq_2, reference_local_seq_1,reference_local_seq_2]

       
    tmp_dir = tempfile.mkdtemp()
    # tmp_dir = "tmp"
    # os.makedirs(tmp_dir)
    # print(tmp_dir)

    temp_key = ""
    temp_id2seq = {}
    temp_junc_seq = ""
    temp_total_read_count = 0
    key2sread = {}
    key2sread_info = {}
    key2sread_count = {}
    key2sread_count_all = {}
    with open(output_file + ".tmp3.long_read_seq.sorted") as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')

            if temp_key != F[0]:
                if temp_key != "":
                    hout2.close()
                    alignment_info_var_1 = ssw_check(tmp_dir + '/' + temp_key + ".variant_seq_1.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
                    alignment_info_var_2 = ssw_check(tmp_dir + '/' + temp_key + ".variant_seq_2.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
                    if is_short_del_dup(temp_key):
                        alignment_info_ref_1 = ssw_check(tmp_dir + '/' + temp_key + ".reference_local_seq_1.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
                        alignment_info_ref_2 = ssw_check(tmp_dir + '/' + temp_key + ".reference_local_seq_2.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)

                    all_rnames = list(set(list(alignment_info_var_1.keys()) + list(alignment_info_var_2.keys())))

                    if is_short_del_dup(temp_key):
                        supporting_reads = [rname for rname in all_rnames if \
                            (alignment_info_var_1[rname][0] > score_ratio_thres * len(variant_seq_1) and \
                            alignment_info_var_1[rname][1] < start_pos_thres * len(variant_seq_1) and \
                            alignment_info_var_1[rname][2] > end_pos_thres * len(variant_seq_1) and \
                            alignment_info_var_1[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres and \
                            alignment_info_var_1[rname][0] >= alignment_info_ref_2[rname][0] + var_ref_margin_thres) or \
                            (alignment_info_var_2[rname][0] > score_ratio_thres * len(variant_seq_2) and \
                            alignment_info_var_2[rname][1] < start_pos_thres * len(variant_seq_2) and \
                            alignment_info_var_2[rname][2] > end_pos_thres * len(variant_seq_2) and \
                            alignment_info_var_2[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres and \
                            alignment_info_var_2[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres)]
                    else:
                        supporting_reads = [rname for rname in all_rnames if \
                            (alignment_info_var_1[rname][0] > score_ratio_thres * len(variant_seq_1) and \
                            alignment_info_var_1[rname][1] < start_pos_thres * len(variant_seq_1) and \
                            alignment_info_var_1[rname][2] > end_pos_thres * len(variant_seq_1)) or \
                            (alignment_info_var_2[rname][0] > score_ratio_thres * len(variant_seq_2) and \
                            alignment_info_var_2[rname][1] < start_pos_thres * len(variant_seq_2) and \
                            alignment_info_var_2[rname][2] > end_pos_thres * len(variant_seq_2))]

                    # filtering by mapping qualities
                    supporting_reads = [rname for rname in supporting_reads if \
                        rname in key2rname2mapq[temp_key] and \
                        key2rname2mapq[temp_key][rname][0] is not None and key2rname2mapq[temp_key][rname][0] >= variant_sread_min_mapq and \
                        key2rname2mapq[temp_key][rname][1] is not None and key2rname2mapq[temp_key][rname][1] >= variant_sread_min_mapq]

                    key2sread[temp_key] = supporting_reads
                    key2sread_info[temp_key] = { rname: alignment_info_var_1[rname] + alignment_info_var_2[rname] for rname in supporting_reads}
                    key2sread_count[temp_key] = len(supporting_reads)
                    key2sread_count_all[temp_key] = len(all_rnames)

                    # print(temp_key + '\t' + str(key2sread_count_all[temp_key]) + '\t' + str(key2sread_count[temp_key]))

                hout2 = open(tmp_dir + '/' + F[0] + ".long_read_seq.fa", 'w')
                temp_key = F[0]
                temp_total_read_count = 0
                FF = temp_key.split(',')
    
                variant_seq_1, variant_seq_2, reference_local_seq_1, reference_local_seq_2 = key2contig[temp_key]
                with open(tmp_dir + '/' + F[0] + ".variant_seq_1.fa", 'w') as hout1:
                    print('>' + F[0] + '\n' + variant_seq_1, file = hout1) 
                with open(tmp_dir + '/' + F[0] + ".variant_seq_2.fa", 'w') as hout1:
                    print('>' + F[0] + '\n' + variant_seq_2, file = hout1)
                with open(tmp_dir + '/' + F[0] + ".reference_local_seq_1.fa", 'w') as hout1:
                    print('>' + F[0] + '\n' + reference_local_seq_1, file = hout1) 
                with open(tmp_dir + '/' + F[0] + ".reference_local_seq_2.fa", 'w') as hout1:
                    print('>' + F[0] + '\n' + reference_local_seq_2, file = hout1)

            print('>' + F[1] + '\n' + F[2], file = hout2)
            temp_total_read_count = temp_total_read_count + 1


        # last transaction
        if temp_key != "":
            hout2.close()

            alignment_info_var_1 = ssw_check(tmp_dir + '/' + temp_key + ".variant_seq_1.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
            alignment_info_var_2 = ssw_check(tmp_dir + '/' + temp_key + ".variant_seq_2.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
            if is_short_del_dup(temp_key):
                alignment_info_ref_1 = ssw_check(tmp_dir + '/' + temp_key + ".reference_local_seq_1.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)
                alignment_info_ref_2 = ssw_check(tmp_dir + '/' + temp_key + ".reference_local_seq_2.fa", tmp_dir + '/' + temp_key + ".long_read_seq.fa", use_ssw_lib)

            # all_rnames = list(set(alignment_info_var_1))
            all_rnames = list(set(list(alignment_info_var_1.keys()) + list(alignment_info_var_2.keys())))

            if is_short_del_dup(temp_key):
                supporting_reads = [rname for rname in all_rnames if \
                    (alignment_info_var_1[rname][0] > score_ratio_thres * len(variant_seq_1) and \
                    alignment_info_var_1[rname][1] < start_pos_thres * len(variant_seq_1) and \
                    alignment_info_var_1[rname][2] > end_pos_thres * len(variant_seq_1) and \
                    alignment_info_var_1[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres and \
                    alignment_info_var_1[rname][0] >= alignment_info_ref_2[rname][0] + var_ref_margin_thres) or \
                    (alignment_info_var_2[rname][0] > score_ratio_thres * len(variant_seq_2) and \
                    alignment_info_var_2[rname][1] < start_pos_thres * len(variant_seq_2) and \
                    alignment_info_var_2[rname][2] > end_pos_thres * len(variant_seq_2) and \
                    alignment_info_var_2[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres and \
                    alignment_info_var_2[rname][0] >= alignment_info_ref_1[rname][0] + var_ref_margin_thres)]
            else:   
                supporting_reads = [rname for rname in all_rnames if \
                    (alignment_info_var_1[rname][0] > score_ratio_thres * len(variant_seq_1) and \
                    alignment_info_var_1[rname][1] < start_pos_thres * len(variant_seq_1) and \
                    alignment_info_var_1[rname][2] > end_pos_thres * len(variant_seq_1)) or \
                    (alignment_info_var_2[rname][0] > score_ratio_thres * len(variant_seq_2) and \
                    alignment_info_var_2[rname][1] < start_pos_thres * len(variant_seq_2) and \
                    alignment_info_var_2[rname][2] > end_pos_thres * len(variant_seq_2))]

            # filtering by mapping qualities
            supporting_reads = [rname for rname in supporting_reads if \
                rname in key2rname2mapq[temp_key] and \
                key2rname2mapq[temp_key][rname][0] is not None and key2rname2mapq[temp_key][rname][0] >= variant_sread_min_mapq and \
                key2rname2mapq[temp_key][rname][1] is not None and key2rname2mapq[temp_key][rname][1] >= variant_sread_min_mapq] 

            key2sread[temp_key] = supporting_reads
            key2sread_info[temp_key] = { rname: alignment_info_var_1[rname] + alignment_info_var_2[rname] for rname in supporting_reads}
            key2sread_count[temp_key] = len(supporting_reads)
            key2sread_count_all[temp_key] = len(all_rnames)


    shutil.rmtree(tmp_dir)
    if not debug: subprocess.call(["rm" ,"-rf", output_file + ".tmp3.long_read_seq.sorted"])

    return([key2sread_count, key2sread_count_all, key2sread_info])



def long_read_validate_main(result_file, tumor_bam, output, sread_file, reference, control_bam, var_read_min_mapq, use_ssw_lib, debug):

    key2sread_count_tumor, key2sread_count_all_tumor, key2sread_info_tumor = long_read_validate_by_alignment(result_file, output, tumor_bam, reference, use_ssw_lib, debug, variant_sread_min_mapq = var_read_min_mapq, score_ratio_thres = 1.2, start_pos_thres = 0.1, end_pos_thres = 0.9, var_ref_margin_thres = 20)

    if control_bam is not None:
        key2sread_count_control, key2sread_count_all_control, key2sread_info_control = long_read_validate_by_alignment(result_file, output, control_bam, reference, use_ssw_lib, debug, variant_sread_min_mapq = var_read_min_mapq, score_ratio_thres = 1.2, start_pos_thres = 0.1, end_pos_thres = 0.9, var_ref_margin_thres = 20)

    hout = open(output, 'w')
    with open(result_file, 'r') as hin:

        for line in hin:
            if line.startswith("#") or line.startswith("Chr_1"): continue

            F = line.rstrip('\n').split('\t')
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6]
            if tinseq == "---": tinseq = ''
            key = ','.join([tchr1, str(tpos1), tdir1, tchr2, str(tpos2), tdir2, str(len(tinseq))])

            sread_count_tumor = key2sread_count_tumor[key] if key in key2sread_count_tumor else 0
            sread_count_all_tumor = key2sread_count_all_tumor[key] if key in key2sread_count_all_tumor else 0
            # sread_id = ','.join(key2sread_tumor[key])

            if control_bam is not None:
                sread_count_control = key2sread_count_control[key] if key in key2sread_count_control else 0
                sread_count_all_control = key2sread_count_all_control[key] if key in key2sread_count_all_control else 0

            if control_bam is not None:
                print('\t'.join(F[:7]) + '\t' + str(sread_count_all_tumor) + '\t' + str(sread_count_tumor) + '\t' + \
                      str(sread_count_all_control) + '\t' + str(sread_count_control), file = hout)
                      # str(sread_count_all_control) + '\t' + str(sread_count_control) + '\t' + sread_id, file = hout)
            else:
                # print('\t'.join(F) + '\t' + str(sread_count_all_tumor) + '\t' + str(sread_count_tumor) + '\t' + sread_id, file = hout)
                print('\t'.join(F[:7]) + '\t' + str(sread_count_all_tumor) + '\t' + str(sread_count_tumor), file = hout)

    hout.close()

    hout = open(sread_file, 'w')
    for key in key2sread_info_tumor:
        for sread in key2sread_info_tumor[key]:
            sinfo = '\t'.join([str(x) for x in key2sread_info_tumor[key][sread]])
            print(key + '\t' + sread + '\t' + sinfo, file = hout)

    hout.close()

if __name__ == "__main__":

    import sys
    result_file = sys.argv[1]
    tumor_bam = sys.argv[2]
    output = sys.argv[3]
    reference = sys.argv[4]
    control_bam = sys.argv[5]

    long_read_validate_main(result_file, tumor_bam, output, output + ".sread.txt", reference, control_bam, False)






#!/usr/bin/env python

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def get_version():
    with open(path.join(here, "nanomonsv/version.py"), encoding = 'utf-8') as hin:
        for line in hin:
            if line.startswith("__version__"):
                version = line.partition('=')[2]
                return version.strip().strip('\'"')
    raise ValueError('Could not find version.')

setup(
    name = 'nanomonsv',
    version = get_version(),
    description='Python tools for detecting structural variation from nanopore sequence data',
    long_description=long_description, 
    long_description_content_type='text/markdown',
    url = 'https://github.com/friend1ws/nanomonsv',
    author = 'Yuichi Shiraishi',
    author_email = 'friend1ws@gamil.com',
    license = 'GPLv3',

    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],

    packages = find_packages(exclude = ['tests', 'LINE_db']),
    package_data={'nanomonsv': ['data/*']},

    # install_requires = ["numpy", "parasail", "pysam"],
    entry_points = {'console_scripts': ['nanomonsv = nanomonsv:main']}

)





#! /usr/bin/env python

import sys

input_file = sys.argv[1]

segments = []
with open(input_file, 'r') as hin:
    for line in hin:
        F = line.rstrip('\n').split('\t')

        segments.append([F[:16] + ['+']])



update = True
while update:
    update = False
    for seg1 in range(len(segments) - 1):
        for seg2 in range(seg1 + 1, len(segments)):

            break1_1 = '\t'.join(segments[seg1][0][:3])
            break1_2 = '\t'.join(segments[seg1][0][4:7])
            break1_3 = '\t'.join(segments[seg1][-1][7:10])
            break1_4 = '\t'.join(segments[seg1][-1][11:14])

            break2_1 = '\t'.join(segments[seg2][0][:3])
            break2_2 = '\t'.join(segments[seg2][0][4:7])
            break2_3 = '\t'.join(segments[seg2][-1][7:10])
            break2_4 = '\t'.join(segments[seg2][-1][11:14])

            if break1_3 == break2_1 and break1_4 == break2_2:
                sdir1, sdir2 = '+', '-'
                update = True
            elif break1_3 == break2_4 and break1_4 == break2_3:
                sdir1, sdir2 = '+', '+'
                update = True
            elif break1_1 == break2_3 and break1_2 == break2_4:
                sdir1, sdir2 = '-', '+'
                update = True
            elif break1_1 == break2_2 and break1_2 == break2_1:
                sdir1, sdir2 = '-', '-'
                update = True

            if update: break

        if update: break


    if update:
        if sdir1 == sdir2:
            for j in range(len(segments[seg2])):
                segments[seg2][j] = segments[seg2][j][11:14] + [segments[seg2][j][10]] + segments[seg2][j][7:10] + \
                                    segments[seg2][j][4:7] + [segments[seg2][j][3]] + segments[seg2][j][:3] + \
                                    segments[seg2][j][14:16] + ['+' if segments[seg2][j][16] == '-' else '-']

                segments[seg2].reverse()

        if sdir1 == '+':
            segments[seg1] = segments[seg1] + segments[seg2]
        else:
            segments[seg1] = segments[seg2] + segments[seg1]

        del(segments[seg2])
        

for seg in segments:

    for s in seg:

        print(s)


    print()
    


"""
8       130933956       +       ---     8       130871365       -       8       130872760       +       ---     8       131401249       +       True    1395 
"""






#! /usr/bin/env python

import csv, sys

sv_get_file = sys.argv[1]
output_file = sys.argv[2]

with open(sv_get_file, 'r') as hin, open(output_file, 'w') as hout:

    dreader = csv.DictReader(hin, delimiter = '\t')

    # print header
    print('\t'.join(dreader.fieldnames) + '\t' + "SV_Type", file = hout)

    for F in dreader:

        sv_type = None
        inseq = '' if F["Inserted_Seq"] == '---' else F["Inserted_Seq"]
        if F["Chr_1"] != F["Chr_2"]:
            sv_type = "Translocation"
        elif F["Dir_1"] == F["Dir_2"]:
            sv_type = "Inversion"
        elif F["Dir_1"] == '-' and F["Dir_2"] == '+':
            sv_type = "Duplication"
        elif F["Dir_1"] == '+' and F["Dir_2"] == '-':
            if len(inseq) > int(F["Pos_2"]) - int(F["Pos_1"]) + 1:
                sv_type = "Insertion"
            else:
                sv_type = "Deletion"

        print('\t'.join(F.values()) + '\t' + sv_type, file = hout)






#! /usr/bin/env python3

import sys, os, subprocess, shutil, random, copy
import pysam
import parasail

from .pyssw import *
from .my_seq import reverse_complement
from .utils import get_alignment_object

# when the total read number exceeds max_read_num, then max_read_num reads are randomly selected
def bam_subsample_fetch(alignment_h, tchr, tstart, tend, max_read_num = 500):

    rec = 0
    for read in alignment_h.fetch(tchr, tstart, tend):
        rec = rec + 1

    selected_inds = random.sample(range(rec), min(max_read_num, rec))

    rec2 = 0
    for read in alignment_h.fetch(tchr, tstart, tend):
        if rec2 in selected_inds:
            yield read
        rec2 = rec2 + 1

# function for gathering sequence read for realignment validation
def gather_local_read_for_realignment(sv_file, alignment_file, output_file, reference_fasta,
    sbnd_file = None, output_file_sbnd = None, validate_sequence_length = 200, check_read_max_num = 500,
    sort_option = None):

    alignment_h = get_alignment_object(alignment_file, reference_fasta)

    rname2key = {}
    key2rname2mapq = {}
    with open(sv_file, 'r') as hin:
        for line in hin:
            if line.startswith("#") or line.startswith("Chr_1"): continue
            F = line.rstrip('\n').split('\t')
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid = F[0], int(F[1]), F[2], F[3], int(F[4]), F[5], F[6], F[7]
            # if tinseq == "---": tinseq = ''
            key = f"{tchr1},{tpos1},{tdir1},{tchr2},{tpos2},{tdir2},{tinseq},{tid}"

            if key not in key2rname2mapq: key2rname2mapq[key] = {}

            for read in bam_subsample_fetch(alignment_h, tchr1, max(tpos1 - 100, 0), tpos1 + 100):
    
                if read.qname not in rname2key: rname2key[read.qname] = []
                rname2key[read.qname].append(key)

                if read.qname not in key2rname2mapq[key]: key2rname2mapq[key][read.qname] = [None, None]
                key2rname2mapq[key][read.qname][0] = read.mapping_quality

            for read in bam_subsample_fetch(alignment_h, tchr2, max(tpos2 - 100, 0), tpos2 + 100):

                if read.qname not in rname2key: rname2key[read.qname] = []
                rname2key[read.qname].append(key)

                if read.qname not in key2rname2mapq[key]: key2rname2mapq[key][read.qname] = [None, None]
                key2rname2mapq[key][read.qname][1] = read.mapping_quality

    # remove duplicated keys
    for rname in rname2key:
        keys = list(set(rname2key[rname]))
        rname2key[rname] = keys

    # for sbnd candidate
    if sbnd_file is not None:
        rname2key_sbnd = {}
        key2rname2mapq_sbnd = {}
        with open(sbnd_file, 'r') as hin:
            for line in hin:
                if line.startswith("#") or line.startswith("Chr_1"): continue
                F = line.rstrip('\n').split('\t')
                tchr, tpos, tdir, tseq, tid = F[0], int(F[1]), F[2], F[3][:validate_sequence_length], F[4]
                key = f"{tchr},{tpos},{tdir},{tseq},{tid}"

                if key not in key2rname2mapq_sbnd: key2rname2mapq_sbnd[key] = {}

                for read in bam_subsample_fetch(alignment_h, tchr, max(tpos - 100, 0), tpos + 100):

                    if read.qname not in rname2key_sbnd: rname2key_sbnd[read.qname] = []
                    rname2key_sbnd[read.qname].append(key)
                    
                    key2rname2mapq_sbnd[key][read.qname] = read.mapping_quality

        # remove duplicated keys
        for rname in rname2key_sbnd:
            keys = list(set(rname2key_sbnd[rname]))
            rname2key_sbnd[rname] = keys

 
    hout = open(output_file + ".tmp.unsorted", 'w')
    if sbnd_file is not None: hout_sbnd = open(output_file_sbnd + ".tmp.unsorted", 'w')
    for read in alignment_h.fetch():

        flags = format(int(read.flag), "#014b")[:1:-1]

        # skip supplementary alignment
        if flags[8] == "1" or flags[11] == "1": continue

        # skip duplicated reads             
        if flags[10] == "1": continue

        if read.qname in rname2key:

            read_seq = read.query_sequence
            if flags[4] == "1": read_seq = reverse_complement(read_seq)

            for key in rname2key[read.qname]:
                mapq1, mapq2 = key2rname2mapq[key][read.qname]
                print(f"{key}\t{read.qname}\t{mapq1}\t{mapq2}\t{read_seq}", file = hout)

        if sbnd_file is not None and read.qname in rname2key_sbnd:

            read_seq = read.query_sequence
            if flags[4] == "1": read_seq = reverse_complement(read_seq)

            for key in rname2key_sbnd[read.qname]:
                mapq = key2rname2mapq_sbnd[key][read.qname]
                print(f"{key}\t{read.qname}\t{mapq}\tNone\t{read_seq}", file = hout_sbnd)

    hout.close()
    if sbnd_file is not None: hout_sbnd.close()

    with open(output_file, 'w') as hout:
        subprocess.call(["sort", "-k1,1"] + sort_option.split(" ") + [output_file + ".tmp.unsorted"], stdout = hout)
    os.remove(output_file + ".tmp.unsorted")

    if sbnd_file is not None:
        with open(output_file_sbnd, 'w') as hout:
            subprocess.call(["sort", "-k1,1"] + sort_option.split(" ") + [output_file_sbnd + ".tmp.unsorted"], stdout = hout)
        os.remove(output_file_sbnd + ".tmp.unsorted")

    alignment_h.close()


def ssw_check(query, target, use_ssw = False):

    user_matrix = parasail.matrix_create("ACGT", 2, -2)

    alignment_info = {}
    for sQId, sQSeq, sQQual in read(query):

        sQSeq_r = reverse_complement(sQSeq)

        for sTId, sTSeq, STQual in read(target):

            res = parasail.ssw(sQSeq, sTSeq, 3, 1, user_matrix)
            res_r = parasail.ssw(sQSeq_r, sTSeq, 3, 1, user_matrix)

            if res.score1 > res_r.score1:
                score = res.score1
                qstart, qend = res.read_begin1 + 1., res.read_end1 + 1
                tstart, tend = res.ref_begin1 + 1, res.ref_end1 + 1
                strand = '+'
            else:
                score = res_r.score1
                qstart, qend = len(sQSeq) - res_r.read_end1, len(sQSeq) - res_r.read_begin1
                tstart, tend = res_r.ref_begin1 + 1, res_r.ref_end1 + 1
                strand = '-'

            alignment_info[sTId] = [score, int(qstart), int(qend), int(tstart), int(tend), strand]

    return(alignment_info)


class Alignment_counter(object):

    def __init__(self, output_count_file, output_alignment_info_file, reference_fasta,
        var_read_min_mapq, score_ratio_thres, use_ssw_lib, debug, validate_sequence_length = 200,  
        start_pos_thres = 0.1, end_pos_thres = 0.9, var_ref_margin_thres = 20):

        self.temp_key = None
        self.temp_key2 = None # here, the insertion sequence is converted to its length. this is to shorten file names.
        self.hout_count = open(output_count_file, 'w')
        self.hout_ainfo = open(output_alignment_info_file, 'w')
        self.reference_fasta_h = pysam.FastaFile(reference_fasta)

        self.reference_segment_1 = None
        self.reference_segment_2 = None
        self.variant_segment_1 = None
        self.variant_segment_2 = None
        self.readid2mapq = {}

        self.validate_sequence_length = validate_sequence_length
        self.score_ratio_thres = score_ratio_thres
        self.start_pos_thres = start_pos_thres
        self.end_pos_thres = end_pos_thres
        self.var_ref_margin_thres = var_ref_margin_thres
        self.var_read_min_mapq = var_read_min_mapq
        self.use_ssw_lib = use_ssw_lib
        self.debug = debug

        self.tmp_dir = output_count_file + ".tmp_dir"
        os.makedirs(self.tmp_dir, exist_ok = True)



    def __del__(self):

        self.hout_count.close()
        self.hout_ainfo.close()
        self.reference_fasta_h.close()
        if not self.debug:
            shutil.rmtree(self.tmp_dir)

    def initialize(self, key, is_sbnd = False):
        self.temp_key = key
        # the insertion sequence is converted to its length. this is to shorten file names.
        tkeys = self.temp_key.split(',')
        tkeys[-2] = '' if tkeys[-2] == '---' else tkeys[-1]
        tkeys[-2] = str(len(tkeys[-2]))
        self.temp_key2 = ','.join(tkeys)

        self.is_inseq = True if tkeys[-2] != '' else False

        if is_sbnd == False:
            self.is_short_del_dup = False
            if tkeys[6] == "---": tkeys[6] = ''
            if tkeys[0] == tkeys[3] and tkeys[2] == '+' and tkeys[5] == '-' and int(tkeys[4]) - int(tkeys[1]) + len(tkeys[6]) < 300:
                self.is_short_del_dup = True
            elif tkeys[0] == tkeys[3] and tkeys[2] == '-' and tkeys[5] == '+' and int(tkeys[4]) - int(tkeys[1]) + len(tkeys[6]) < 300:
                self.is_short_del_dup = True

        self.readid2mapq = {}
        self.temp_long_read_seq_file_h = open(self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", 'w')
        if is_sbnd:
            self.generate_segment_fasta_files_sbnd()
        else:
            self.generate_segment_fasta_files()

    def generate_segment_fasta_files(self):

        tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid = self.temp_key.split(',')
        tpos1, tpos2 = int(tpos1), int(tpos2)
        tinseq = '' if tinseq == "---" else tinseq

        # reference_local_seq
        self.reference_segment_1 = self.reference_fasta_h.fetch(tchr1, max(tpos1 - self.validate_sequence_length - 1, 0), tpos1 + self.validate_sequence_length - 1).upper()
        self.reference_segment_2 = self.reference_fasta_h.fetch(tchr2, max(tpos2 - self.validate_sequence_length - 1, 0), tpos2 + self.validate_sequence_length - 1).upper()

        # variant_seq
        variant_seq = ""
        if tdir1 == '+':
            tseq = self.reference_fasta_h.fetch(tchr1, max(tpos1 - self.validate_sequence_length - 1, 0), tpos1 - 1).upper()
        else:
            tseq = self.reference_fasta_h.fetch(tchr1, tpos1 - 1, tpos1 + self.validate_sequence_length - 1).upper()
            tseq = reverse_complement(tseq)

        if tdir1 == "+":
            variant_seq = tseq + tinseq
        else:
            variant_seq = tseq + reverse_complement(tinseq)

        if tdir2 == '-':
            tseq = self.reference_fasta_h.fetch(tchr2, tpos2 - 1, tpos2 + self.validate_sequence_length - 1).upper()
        else:
            tseq = self.reference_fasta_h.fetch(tchr2, max(tpos2 - self.validate_sequence_length - 1, 0), tpos2 - 1).upper()
            tseq = reverse_complement(tseq)

        variant_seq = variant_seq + tseq

        self.variant_segment_1 = variant_seq[:min(2 * self.validate_sequence_length, len(variant_seq))]
        self.variant_segment_2 = variant_seq[-min(2 * self.validate_sequence_length, len(variant_seq)):]


    def generate_segment_fasta_files_sbnd(self):

        tchr, tpos, tdir, tcontig, tid = self.temp_key.split(',')
        tpos = int(tpos)

        # referene_local_seq
        self.reference_segment_1 = self.reference_fasta_h.fetch(tchr, max(tpos - self.validate_sequence_length - 1, 0), tpos + self.validate_sequence_length - 1).upper()

        # variant_seq
        variant_seq = ''
        if tdir == '+':
            tseq = self.reference_fasta_h.fetch(tchr, max(tpos - self.validate_sequence_length - 1, 0), tpos).upper()
            tseq = tseq + tcontig
        else:
            tseq = self.reference_fasta_h.fetch(tchr, tpos - 1, tpos + self.validate_sequence_length - 1).upper()
            tseq = reverse_complement(tseq)
        self.variant_segment_1 = tseq + tcontig


    def add_long_read_seq(self, treadid, tseq, tmapq1, tmapq2):

        self.readid2mapq[treadid] = [int(tmapq1) if tmapq1 != "None" else None, int(tmapq2) if tmapq2 != "None" else None]
        print(f">{treadid}\n{tseq}", file = self.temp_long_read_seq_file_h)


    def count_alignment_and_print(self):

        # close file hundle for writing long_read_seq.fa if it is open
        if self.temp_long_read_seq_file_h is not None: self.temp_long_read_seq_file_h.close()

        with open(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_1.fa", 'w') as hout:
            print('>' + self.temp_key2 + '\n' + self.variant_segment_1, file = hout)
        with open(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_2.fa", 'w') as hout:
            print('>' + self.temp_key2 + '\n' + self.variant_segment_2, file = hout)

        alignment_info_var_1 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_1.fa",
            self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)

        if self.is_inseq:
            alignment_info_var_2 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_2.fa",
                self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)
        else:
            alignment_info_var_2 = copy.copy(alignment_info_var_1)


        if self.is_short_del_dup:
            with open(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_1.fa", 'w') as hout:
                print('>' + self.temp_key2 + '\n' + self.reference_segment_1, file = hout)

            with open(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_2.fa", 'w') as hout:
                print('>' + self.temp_key2 + '\n' + self.reference_segment_2, file = hout)

            alignment_info_ref_1 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_1.fa",
                self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)
            alignment_info_ref_2 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_2.fa",
                self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)


        all_rnames = list(set(list(alignment_info_var_1.keys()) + list(alignment_info_var_2.keys())))

        supporting_reads = [rname for rname in all_rnames if \
            (alignment_info_var_1[rname][0] > self.score_ratio_thres * len(self.variant_segment_1) and \
            alignment_info_var_1[rname][1] < self.start_pos_thres * len(self.variant_segment_1) and \
            alignment_info_var_1[rname][2] > self.end_pos_thres * len(self.variant_segment_1)) or \
            (alignment_info_var_2[rname][0] > self.score_ratio_thres * len(self.variant_segment_2) and \
            alignment_info_var_2[rname][1] < self.start_pos_thres * len(self.variant_segment_2) and \
            alignment_info_var_2[rname][2] > self.end_pos_thres * len(self.variant_segment_2))]

        # for short deletion or insertion
        if self.is_short_del_dup:
            supporting_reads = [rname for rname in supporting_reads if \
                (alignment_info_var_1[rname][0] >= alignment_info_ref_1[rname][0] + self.var_ref_margin_thres and \
                alignment_info_var_1[rname][0] >= alignment_info_ref_2[rname][0] + self.var_ref_margin_thres) or \
                (alignment_info_var_2[rname][0] >= alignment_info_ref_1[rname][0] + self.var_ref_margin_thres and \
                alignment_info_var_2[rname][0] >= alignment_info_ref_2[rname][0] + self.var_ref_margin_thres)]

        # filtering by mapping qualities
        supporting_reads = [rname for rname in supporting_reads if \
            self.readid2mapq[rname][0] is not None and self.readid2mapq[rname][0] >= self.var_read_min_mapq and \
            self.readid2mapq[rname][1] is not None and self.readid2mapq[rname][1] >= self.var_read_min_mapq]


        tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, tid = self.temp_key.split(',')
        print(f"{tchr1}\t{tpos1}\t{tdir1}\t{tchr2}\t{tpos2}\t{tdir2}\t{tinseq}\t{tid}\t{len(all_rnames)}\t{len(supporting_reads)}", file = self.hout_count)
        
        sread_info = { rname: alignment_info_var_1[rname] + alignment_info_var_2[rname] for rname in supporting_reads}
        for rname in supporting_reads:
            if self.is_short_del_dup:
                sinfo = '\t'.join([str(x) for x in alignment_info_var_1[rname] + alignment_info_var_2[rname] + alignment_info_ref_1[rname] + alignment_info_ref_2[rname]])
            else:
                sinfo = '\t'.join([str(x) for x in alignment_info_var_1[rname] + alignment_info_var_2[rname] + ["None"] * 12])
            print(f"{tchr1}\t{tpos1}\t{tdir1}\t{tchr2}\t{tpos2}\t{tdir2}\t{tinseq}\t{tid}\t{rname}\t{sinfo}", file = self.hout_ainfo)


    def count_alignment_and_print_sbnd(self):

        # close file hundle for writing long_read_seq.fa if it is open
        if self.temp_long_read_seq_file_h is not None: self.temp_long_read_seq_file_h.close()

        with open(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_1.fa", 'w') as hout:
            print('>' + self.temp_key2 + '\n' + self.variant_segment_1, file = hout)
        with open(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_1.fa", 'w') as hout:
            print('>' + self.temp_key2 + '\n' + self.reference_segment_1, file = hout)

        alignment_info_var_1 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".variant_seg_1.fa",
            self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)

        alignment_info_ref_1 = ssw_check(self.tmp_dir + '/' + self.temp_key2 + ".reference_seg_1.fa",
            self.tmp_dir + '/' + self.temp_key2 + ".long_read_seq.fa", self.use_ssw_lib)

        all_rnames = list(set(list(alignment_info_var_1.keys())))

        supporting_reads = [rname for rname in all_rnames if \
            (alignment_info_var_1[rname][0] > self.score_ratio_thres * len(self.variant_segment_1) and \
            alignment_info_var_1[rname][1] < self.start_pos_thres * len(self.variant_segment_1) and \
            alignment_info_var_1[rname][2] > self.end_pos_thres * len(self.variant_segment_1))]

        supporting_reads = [rname for rname in supporting_reads if \
            (alignment_info_var_1[rname][0] >= alignment_info_ref_1[rname][0] + self.var_ref_margin_thres)]

        # filtering by mapping qualities
        supporting_reads = [rname for rname in supporting_reads if \
            self.readid2mapq[rname][0] is not None and self.readid2mapq[rname][0] >= self.var_read_min_mapq]

        tchr, tpos, tdir, tcontig, tid = self.temp_key.split(',')
        print(f"{tchr}\t{tpos}\t{tdir}\t{tcontig}\t{tid}\t{len(all_rnames)}\t{len(supporting_reads)}", file = self.hout_count)

        sread_info = { rname: alignment_info_var_1[rname] for rname in supporting_reads}
        for rname in supporting_reads:
            sinfo = '\t'.join([str(x) for x in alignment_info_var_1[rname] ])
            print(f"{tchr}\t{tpos}\t{tdir}\t{tcontig}\t{tid}\t{rname}\t{sinfo}", file = self.hout_ainfo)



def count_sread_by_alignment(sv_file, alignment_file, output_count_file, output_alignment_info_file, reference_fasta, 
    sbnd_file = None, output_count_file_sbnd = None, output_alignment_info_file_sbnd = None,
    check_read_max_num = 500, var_read_min_mapq = 0, score_ratio_thres = 1.2, use_ssw_lib = False, sort_option = None, debug = False):

    gather_local_read_for_realignment(sv_file, alignment_file, output_count_file + ".tmp.local_read_for_realignment", reference_fasta,
        sbnd_file = sbnd_file, 
        output_file_sbnd = output_count_file_sbnd + ".tmp.local_read_for_realignment" if output_count_file_sbnd is not None else None,
        check_read_max_num = check_read_max_num,
        sort_option = sort_option)

    alignment_counter = Alignment_counter(output_count_file, output_alignment_info_file, reference_fasta,
        var_read_min_mapq, score_ratio_thres, use_ssw_lib, debug)

    with open(output_count_file + ".tmp.local_read_for_realignment", 'r') as hin:
        for line in hin:
            tkey, treadid, tmapq1, tmpaq2, tseq = line.rstrip('\n').split('\t')
            if alignment_counter.temp_key != tkey:
                if alignment_counter.temp_key is not None:
                    alignment_counter.count_alignment_and_print()
                alignment_counter.initialize(tkey)

            alignment_counter.add_long_read_seq(treadid, tseq, tmapq1, tmpaq2)

        if alignment_counter.temp_key is not None:
            alignment_counter.count_alignment_and_print()


    if not debug:
        os.remove(output_count_file + ".tmp.local_read_for_realignment")

    if sbnd_file is None: return

    alignment_counter_sbnd = Alignment_counter(output_count_file_sbnd, output_alignment_info_file_sbnd, reference_fasta,
        var_read_min_mapq, score_ratio_thres, use_ssw_lib, debug)

    with open(output_count_file_sbnd + ".tmp.local_read_for_realignment", 'r') as hin:
        for line in hin:
            tkey, treadid, tmapq1, tmapq2, tseq = line.rstrip('\n').split('\t')
            tmapq2 = "None"
            if alignment_counter_sbnd.temp_key != tkey:
                if alignment_counter_sbnd.temp_key is not None:
                    alignment_counter_sbnd.count_alignment_and_print_sbnd()
                alignment_counter_sbnd.initialize(tkey, is_sbnd = True)

            alignment_counter_sbnd.add_long_read_seq(treadid, tseq, tmapq1, tmapq2)

        if alignment_counter_sbnd.temp_key is not None:
            alignment_counter_sbnd.count_alignment_and_print_sbnd()

    if not debug:
        os.remove(output_count_file_sbnd + ".tmp.local_read_for_realignment")




#! /usr/bin/env python3

import numpy as np

def sw_jump(contig, region1_seq, region2_seq, match_score = 1, mismatch_penalty = 2, gap_cost = 2, jump_cost = 2, minimum_score = 10):

    H1 = np.zeros((len(contig) + 1, len(region1_seq) + 1), np.int32)
    H1_path = np.zeros((len(contig) + 1, len(region1_seq) + 1), np.int32)
    H2 = np.zeros((len(contig) + 1, len(region2_seq) + 1), np.int32)
    H2_path = np.zeros((len(contig) + 1, len(region2_seq) + 1), np.int32)
    H2_origin_i = np.zeros((len(contig) + 1, len(region2_seq) + 1), np.int32)
    H2_origin_j = np.zeros((len(contig) + 1, len(region2_seq) + 1), np.int32)
    H2_origin_score = np.zeros((len(contig) + 1, len(region2_seq) + 1), np.int32)

    
    for i in range(1, H1.shape[0]):
        for j in range(1, H1.shape[1]):
            match = H1[i - 1, j - 1] + (match_score if contig[i - 1] == region1_seq[j - 1] else - mismatch_penalty)
            delete = H1[i - 1, j] - gap_cost
            insert = H1[i, j - 1] - gap_cost
            tscores = (float("-inf"), match, delete, insert)
            H1[i, j] = max(tscores)
            H1_path[i,j] = tscores.index(max(tscores))

    H1_max = np.amax(H1, axis = 1)
    H1_argmax = np.argmax(H1, axis = 1)

    for i in range(1, H2.shape[0]):
        for j in range(1, H2.shape[1]):
            match = H2[i - 1, j - 1] + (match_score if contig[i - 1] == region2_seq[j - 1] else - mismatch_penalty)
            delete = H2[i - 1, j] - gap_cost
            insert = H2[i, j - 1] - gap_cost

            jump = float("-inf")
            org_i = np.argmax(H1_max[:i])
            if H1_max[org_i] >= minimum_score:

                jump = H1_max[org_i] + (match_score if contig[i - 1] == region2_seq[j - 1] else - mismatch_penalty)
                jump_diff = i - org_i
                jump = jump - jump_cost * np.log(jump_diff)
 
            tscores = (float("-inf"), match, delete, insert, jump)
            H2[i, j] = max(tscores)
            H2_path[i, j] = tscores.index(max(tscores))

            if H2_path[i, j] == 4:
                H2_origin_i[i, j] = np.argmax(H1_max[:i])
                H2_origin_j[i, j]  = H1_argmax[H2_origin_i[i, j]]
                H2_origin_score[i, j] = H1_max[org_i]

    # i_cur, j2_cur = np.unravel_index(H2.argmax(), H2.shape)
    # a_string, b_string = a[i_end - 1], b[j_end - 1]
    if np.max(H2[H2.shape[0] - 1, :]) >= np.max(H2[:, H2.shape[1] - 1]):
        j2_cur = np.unravel_index(H2[H2.shape[0] - 1, :].argmax(), (1, H2.shape[1]))[1]
        i_cur = H2.shape[0] - 1
    else:
        j2_cur = H2.shape[1] - 1 
        i_cur = np.unravel_index(H2[:, H2.shape[1] - 1].argmax(), (H2.shape[0], 1))[0]


    contig_match, region1_seq_match, region2_seq_match = '', '', ''
    i_end, i2_end, j2_end = i_cur, i_cur, j2_cur # one-based position

    match_count, deletion_count, insertion_count, jump_count = 0, 0, 0, 0
    H1_score = 0
    scores = []

    while i_cur > 0 and j2_cur > 0:
        scores.append(H2[i_cur, j2_cur])
        if H2_path[i_cur, j2_cur] == 1:
            contig_match, region2_seq_match = contig[i_cur - 1] + contig_match, region2_seq[j2_cur - 1] + region2_seq_match
            i_cur, j2_cur = i_cur - 1, j2_cur - 1
            match_count = match_count + 1
        elif H2_path[i_cur, j2_cur] == 2:
            contig_match, region2_seq_match = contig[i_cur - 1] + contig_match, '-' + region2_seq_match
            i_cur, j2_cur = i_cur - 1, j2_cur
            deletion_count = deletion_count + 1
        elif H2_path[i_cur, j2_cur] == 3: 
            contig_match, region2_seq_match = '-' + contig_match, region2_seq[j2_cur - 1] + region2_seq_match
            i_cur, j2_cur = i_cur, j2_cur - 1
            insertion_count = insertion_count + 1
        elif H2_path[i_cur, j2_cur] ==  4:
            contig_match, region2_seq_match = contig[i_cur - 1] + contig_match, region2_seq[j2_cur - 1] + region2_seq_match
            i1_end = H2_origin_i[i_cur, j2_cur]
            j1_end = H2_origin_j[i_cur, j2_cur]
            H1_score = H2_origin_score[i_cur, j2_cur]
            i2_start = i_cur
            j2_start = j2_cur # one-based alignment start of second fragment
            """
            i2_start, i1_end = i_cur, i_cur - 1
            j2_start = j2_cur # one-based alignment start of second fragment
            i_cur = i_cur - 1
            """
            jump_count = jump_count + 1
            break
        elif H2_path[i_cur, j2_cur] == 0:
            break

    if H1_score < minimum_score or H2.max() - H1_score < minimum_score or jump_count == 0:
        return(None)

    contig_match = ' ' + contig_match
    # j1_end = H1_argmax[i_cur]
    j1_cur = j1_end
    i_cur = i1_end

    while i_cur > 0 and j1_cur > 0:
        scores.append(H1[i_cur, j1_cur])
        if H1_path[i_cur, j1_cur] == 1:
            contig_match, region1_seq_match = contig[i_cur - 1] + contig_match, region1_seq[j1_cur - 1] + region1_seq_match 
            i_cur, j1_cur = i_cur - 1, j1_cur - 1
            match_count = match_count + 1
        elif H1_path[i_cur, j1_cur] == 2:
            contig_match, region1_seq_match = contig[i_cur - 1] + contig_match, '-' + region1_seq_match
            i_cur, j1_cur = i_cur - 1, j1_cur
            deletion_count = deletion_count + 1
        elif H1_path[i_cur, j1_cur] == 3:
            contig_match, region1_seq_match = '-' + contig_match, region1_seq[j1_cur - 1] + region1_seq_match
            i_cur, j1_cur = i_cur, j1_cur - 1
            insertion_count = insertion_count + 1
        elif H1_path[i_cur, j1_cur] == 0:
            break

    i1_start = i_cur
    j1_start = j1_cur + 1

    return(H2.max(), (i1_start, i1_end, i2_start, i2_end), (j1_start, j1_end), (j2_start, j2_end), contig_match, region1_seq_match + ' ' + region2_seq_match)



if __name__ == "__main__":

    A = "AGGAGGGGGTTTCAGATTCCCCCATCCAAAGGAGGTTTTGGCCCCATGGGGAATGAAACAAAGACACATGAACACAGCTGCAAATATGCCCCACCTCTCCCTGCCCCCCCAACTTGCCGGGCCAATGCACCTGCAGTCCCTCACCCTTCTCAGTTCATACCTACTCCCCAGCTCTCTCACTCTTGTATTCTAGATCCCTCAGTACACCAGCGCCCGGGAAGTCAGGGCCGGGCCAGGACTTTGACGTGCTGGTGGCAGACTGAAGGCCAACTCCCATGGGGAGTCCCCAAGGAGAAGGTGGAATCAGGGAAAAAAGGTGAATGGATGACTCAGTCACAGATGGAGCCACAAAGATGAAGCCATTCATTCCACCCACCACCACCACCACAAAAAAAGAACATCTGAGCCCCCCATGAGTCAGGGATGAGCTGGTGCACCTGCCCTCAGGGGCAGCCAATAGAAAAGTGCACCCCTAATGATCTGAAGTAACTTGAAATCACCAGTCATCCTACCCAAAATGCAAACTCAGTGGTTTCCCTCACCTACTCAAACCCTTTAAGGATGCAAAGCACATCTGGTTTTGCTTGCATT"

    B1 = "CAAAGGGGAGTCTCCCAAGGAGAAGAGCCCAGGGCGCAAGGAGCAA"
    B2 = "TTACAATATTATCTGATGCTGAAAGGTGGGGATCAGGGAAAAG"

    score, region1, region2, region3, A_match, B_match = sw_jump(A, B1, B2)

    print(score)
    print(region1)
    print(region2)
    print(region3)
    print(A_match)
    print(B_match)





#! /usr/bin/env python3

import sys, gzip, statistics, logging
import pysam

from .logger import get_logger

logger = get_logger(__name__)

class Sv_cluster(object):

    def __init__(self, tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2,
        treadid, tsize, tinfo1, tinfo2):

        self.chr1 = tchr1
        self.start1 = tstart1
        self.end1 = tend1
        self.dir1 = tdir1
        self.chr2 = tchr2
        self.start2 = tstart2
        self.end2 = tend2
        self.dir2 = tdir2
        self.readids = [treadid]
        self.size = [tsize]
        self.info1 = [tinfo1]
        self.info2 = [tinfo2]


class Sv_clusterer(object):

    def __init__(self, svtype, output_file, control_junction_bedpe = None, control_panel_junction_bedpe = None, bp_bed = None,
        cluster_margin_size = None, size_margin_ratio = None, maximum_local_variant_num = None, skip_margin = None,
        read_num_thres = None, median_mapQ_thres = None, 
        max_overhang_size_thres = None, max_control_read_num = None, control_check_margin = None,
        max_panel_read_num = None, max_panel_sample_num = None):

        self.hout = open(output_file, 'w')
        self.control_tb = None
        if control_junction_bedpe is not None:
            self.control_tb = pysam.TabixFile(control_junction_bedpe)

        self.control_panel_tb = None
        if control_panel_junction_bedpe is not None:
            self.control_panel_tb = pysam.TabixFile(control_panel_junction_bedpe)

        self.bp_tb = None
        if bp_bed is not None:
            self.bp_tb = pysam.TabixFile(bp_bed)
        self.sv_cluster_list = []
        self.svtype = svtype
        self.control_junction_bedpe = control_junction_bedpe
        self.control_panel_junction_bedpe = control_panel_junction_bedpe
        self.bp_bed = bp_bed
        self.cluster_margin_size = cluster_margin_size
        self.size_margin_ratio = size_margin_ratio
        self.maximum_local_variant_num = maximum_local_variant_num
        self.skip_margin = skip_margin
        self.read_num_thres = read_num_thres
        self.median_mapQ_thres = median_mapQ_thres
        self.max_overhang_size_thres = max_overhang_size_thres
        self.max_control_read_num = max_control_read_num
        self.control_check_margin = control_check_margin
        self.max_panel_read_num = max_panel_read_num
        self.max_panel_sample_num = max_panel_sample_num

        self.next_pos_after_skip = 0
        self.temp_chr = None
        self.skipped_pos_list = []

    def __del__(self):
        if len(self.skipped_pos_list) > 0:
            skipped_pls_list_line = ' '.join(self.skipped_pos_list)
            logger.debug(f"Because of too many local variants, clustering procedures were skipped around: {skipped_pls_list_line}")

        self.hout.close()
        if self.control_tb is not None: self.control_tb.close()
        if self.control_panel_tb is not None: self.control_panel_tb.close()
        if self.bp_tb is not None: self.bp_tb.close()

 
    def check_mergeability(self, tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2, 
        treadid, tsize, tinfo1, tinfo2):
   
        for cluster in self.sv_cluster_list:

            if treadid in cluster.readids: continue

            # pairs of chromosome and direction should be the same.
            # two pairs of start and end should overlap.   
            is_mergeable = False
            if self.svtype == "rearrangement": 
                if tchr1 == cluster.chr1 and tchr2 == cluster.chr2 and tdir1 == cluster.dir1 and tdir2 == cluster.dir2 and \
                    tend1 >= cluster.start1 and tstart1 <= cluster.end1 and tend2 >= cluster.start2 and tstart2 <= cluster.end2: 
                    is_mergeable = True

            elif self.svtype in ["insertion", "deletion"]:

                if tchr1 == cluster.chr1 and tstart1 <= cluster.end1 and tend1 >= cluster.start1 and \
                    tstart2 <= cluster.end2 and tend2 >= cluster.start1 and \
                    tsize > (1.0 - self.size_margin_ratio) * float(min(cluster.size)) and \
                    tsize < (1.0 + self.size_margin_ratio) * float(max(cluster.size)):
                    is_mergeable = True
                   
            if not is_mergeable: continue

            cluster.start1 = min(tstart1, cluster.start1)
            cluster.end1 = max(tend1, cluster.end1)
            cluster.start2 = min(tstart2, cluster.start2)
            cluster.end2 = max(tend2, cluster.end2)

            cluster.readids.append(treadid)
            cluster.size.append(tsize)
            cluster.info1.append(tinfo1)
            cluster.info2.append(tinfo2)
                        
            return cluster
            
        return None


    def add_new_cluster(self, tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2, 
        treadid, tsize, tinfo1, tinfo2):

        self.sv_cluster_list.append(
            Sv_cluster(tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2,
                treadid, tsize, tinfo1, tinfo2))


    def filter_rearrangement_cluster(self, cl):
        
        is_filter = False
        if len(cl.readids) < self.read_num_thres: is_filter = True

        median_mapQ1 = statistics.median([int(x.split(',')[5]) for x in cl.info1])
        median_mapQ2 = statistics.median([int(x.split(',')[5]) for x in cl.info2])
        if median_mapQ1 < self.median_mapQ_thres or median_mapQ2 < self.median_mapQ_thres: is_filter = True

        max_overhang_size1 = max([abs(int(x.split(',')[2]) - int(x.split(',')[0])) for x in cl.info1])
        max_overhang_size2 = max([abs(int(x.split(',')[2]) - int(x.split(',')[0])) for x in cl.info2]) 
        if max_overhang_size1 < self.max_overhang_size_thres or max_overhang_size2 < self.max_overhang_size_thres: is_filter = True

        if is_filter == True: return(True)

        control_flag = False
        control_panel_flag = False
        is_short_deletion = False
        if cl.chr1 == cl.chr2 and cl.dir1 == '+' and cl.dir2 == '-' and cl.end2 - cl.start1 <= 500:
            is_short_deletion = True

        if self.control_tb is not None:

            tabix_error_flag = False
            try:
                records = self.control_tb.fetch(cl.chr1, max(0, cl.start1 - 200), cl.end1 + 200)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag = True

            support_read_num = 0
            if not tabix_error_flag:
                for record_line in records:
                    rec = record_line.split('\t')

                    if cl.chr1 != rec[0] or cl.chr2 != rec[3] or cl.dir1 != rec[8] or cl.dir2 != rec[9]: continue
        
                    if is_short_deletion:
                        if cl.start1 <= int(rec[5]) and cl.end2 >= int(rec[1]):
                            support_read_num = support_read_num + 1
                    else:
                        if cl.end1 >= int(rec[1]) - self.control_check_margin and \
                            cl.start1 <= int(rec[2]) + self.control_check_margin and \
                            cl.end2 >= int(rec[4]) - self.control_check_margin and \
                            cl.start2 <= int(rec[5]) + self.control_check_margin:
                            support_read_num = support_read_num + 1

                if support_read_num > self.max_control_read_num:
                    control_flag = True


        if self.control_panel_tb is not None:

            tabix_error_flag = False
            try:
                records = self.control_panel_tb.fetch(cl.chr1, max(0, cl.start1 - 200), cl.end1 + 200)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag = True

            sample2readnum_panel = {}
            if not tabix_error_flag:
                for record_line in records:
                    rec = record_line.split('\t')

                    if cl.chr1 != rec[0] or cl.chr2 != rec[3] or cl.dir1 != rec[8] or cl.dir2 != rec[9]: continue

                    panel_key_match = False
                    if is_short_deletion:
                        if cl.start1 <= int(rec[5]) and cl.end2 >= int(rec[1]):
                            panel_key_match = True
                    else:
                        if cl.end1 >= int(rec[1]) - self.control_check_margin and \
                            cl.start1 <= int(rec[2]) + self.control_check_margin and \
                            cl.end2 >= int(rec[4]) - self.control_check_margin and \
                            cl.start2 <= int(rec[5]) + self.control_check_margin:

                            panel_key_match = True

                    if panel_key_match:
                        readnums = rec[10].split(',')
                        psamples = rec[11].split(',')
                        for psample, readnum in zip(psamples, readnums):
                            if psample not in sample2readnum_panel: 
                                sample2readnum_panel[psample] = 0
                            sample2readnum_panel[psample] = sample2readnum_panel[psample] + int(readnum)

                readnums_panel = [sample2readnum_panel[x] for x in sample2readnum_panel]
                if len([x for x in readnums_panel if x > self.max_panel_read_num]) > self.max_panel_sample_num:
                    control_panel_flag = True

        if control_flag == True or control_panel_flag == True: is_filter = True 

        return(is_filter)

    
    def filter_indel_cluster(self, cl, matched_control_margin = 0):

        is_filter = False
        # add breakpoint info
        if self.bp_tb is not None:
            tabix_error_flag1 = False
            try:
                records = self.bp_tb.fetch(cl.chr1, max(0, cl.start1), cl.end2)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag1 = True

            if not tabix_error_flag1:
                for record_line in records:
                    rec = record_line.split('\t')
                    if cl.chr1 == rec[0] and cl.start1 <= int(rec[2]) and int(rec[2]) <= cl.end1 and rec[5] == '+' or \
                        cl.chr1 == rec[0] and cl.start2 <= int(rec[2]) and int(rec[2]) <= cl.end2 and rec[5] == '-':
    
                        if rec[3] in cl.readids: continue
                        cl.readids.append(rec[3])
                        cl.size.append(rec[5])
                        cl.info1.append(rec[6])

        if len(cl.readids) < self.read_num_thres: is_filter = True

        median_mapQ1 = statistics.median([int(x.split(',')[5]) for x in cl.info1])
        if median_mapQ1 < self.median_mapQ_thres: is_filter = True

        # supporting read number of non secondary alignments
        non_secondary_readnum = len([x.split(',')[10] for x in cl.info1 if x.split(',')[10] == "False"])
        if non_secondary_readnum < self.read_num_thres: is_filter = True

        median_size = statistics.median([int(x) for x in cl.size if x not in ['-', '+']])

        # for indels whose size is below 300, there should be sufficient non-soft-clipping support reads
        if median_size < 300 and len([int(x) for x in cl.size if x not in ['-', '+']]) < self.read_num_thres: 
            is_filter = True
        
        if is_filter == True: return(True)




        control_flag = False
        control_panel_flag = False
        if self.control_tb is not None:

            control_read_num = 0
            tabix_error_flag2 = False
            try:
                records = self.control_tb.fetch(cl.chr1, max(0, cl.start1 - matched_control_margin), cl.end2 + matched_control_margin)
            except Exception as e: 
                logger.debug(f'{e}')
                tabix_error_flag2 = True

            if not tabix_error_flag2:
                for record_line in records:
                    rec = record_line.split('\t')

                    """
                    if cl.chr1 == rec[0] and cl.start1 - self.control_check_margin <= int(rec[2]) and \
                        cl.end2 + self.control_check_margin >= int(rec[1]) and \
                        int(rec[4]) >= median_size * 0.5:
                    """
                    
                    """
                    if cl.chr1 == rec[0] and cl.start1 <= int(rec[1]) <= cl.end1 and \
                        cl.start2 <= int(rec[2]) <= cl.end2 and \
                        median_size * 0.75 <= int(rec[4]) <= median_size * 1.25:
                    """
                    if int(rec[4]) >= median_size * 0.5:
                        control_read_num = control_read_num + 1

            if control_read_num > self.max_control_read_num or float(control_read_num) / len(cl.readids) >= 0.2:
                control_flag = True


        if self.control_panel_tb is not None:
    
            tabix_error_flag2 = False
            try:
                records = self.control_panel_tb.fetch(cl.chr1, max(0, cl.start1 - 50), cl.end2 + 50)
            except Exception as e:
                logger.debug(f'{e}')
                tabix_error_flag2 = True

            sample2readnum_panel = {}
            if not tabix_error_flag2:
                for record_line in records:
                    rec = record_line.split('\t')

                    """
                    if cl.chr1 == rec[0] and cl.start1 - self.control_check_margin <= int(rec[2]) and \
                        cl.end2 + self.control_check_margin >= int(rec[1]) and \
                        int(rec[4]) >= median_size * 0.5:
                    """
                    if cl.chr1 == rec[0] and cl.start1 <= int(rec[1]) <= cl.end1 and \
                        cl.start2 <= int(rec[2]) <= cl.end2 and \
                        median_size * 0.75 <= int(rec[4]) <= median_size * 1.25:

                        readnums = rec[6].split(',')
                        psamples = rec[7].split(',')
                        for psample, readnum in zip(psamples, readnums):
                            if psample not in sample2readnum_panel: sample2readnum_panel[psample] = 0
                            sample2readnum_panel[psample] = sample2readnum_panel[psample] + int(readnum)

                readnums_panel = [sample2readnum_panel[x] for x in sample2readnum_panel]
                if len([x for x in readnums_panel if x > self.max_panel_read_num]) > self.max_panel_sample_num:
                    control_panel_flag = True

        if control_flag == True or control_panel_flag == True: is_filter = True

        return(is_filter)


    def flush_sv_cluster_list(self, current_chr, current_pos):

        remove_cluster = []
        for i in range(len(self.sv_cluster_list)):
            cl = self.sv_cluster_list[i]
            if current_chr != cl.chr1 or current_pos > cl.end1 + self.cluster_margin_size:
                remove_cluster.append(cl)

                if self.svtype == "rearrangement":

                    if self.filter_rearrangement_cluster(cl): continue
                    # print_line_readids = ';'.join(cl.readids)
                    print_line_readids = repr(cl.readids)
                    print_line_info1 = ';'.join(cl.info1)
                    print_line_info2 = ';'.join(cl.info2)
        
                    print(f'{cl.chr1}\t{cl.start1}\t{cl.end1}\t{cl.chr2}\t{cl.start2}\t{cl.end2}\t' +
                        f'{print_line_readids}\t0\t{cl.dir1}\t{cl.dir2}\t{print_line_info1}\t{print_line_info2}',
                        file = self.hout)

                elif self.svtype in ["insertion", "deletion"]:

                    matched_control_margin_choice = 100 if self.svtype == "insertion" else 10
                    if self.filter_indel_cluster(cl, matched_control_margin = matched_control_margin_choice): continue
                    # print_line_readids = ';'.join(cl.readids)
                    print_line_readids = repr(cl.readids)
                    print_line_size = ';'.join([str(x) for x in cl.size])
                    print_line_info1 = ';'.join(cl.info1)

                    print(f'{cl.chr1}\t{cl.start1}\t{cl.end1}\t{cl.chr2}\t{cl.start2}\t{cl.end2}\t' +
                        f'{print_line_readids}\t0\t{cl.dir1}\t{cl.dir2}\t{print_line_size}\t{print_line_info1}',
                        file = self.hout)
                        
        for cl in remove_cluster:
            self.sv_cluster_list.remove(cl)


    def check_exceeding_local_variant_num(self, current_chr, current_pos):

        if len(self.sv_cluster_list) > self.maximum_local_variant_num:
            self.skipped_pos_list.append(current_chr + ':' + str(current_pos))
            self.sv_cluster_list = []
            self.next_pos_after_skip = current_pos + self.skip_margin


def cluster_supporting_reads(input_file, output_file, svtype, control_junction_bedpe = None, 
    control_panel_junction_bedpe = None, bp_bed = None, 
    cluster_margin_size = 100, indel_cluster_margin_size = 10, size_margin_ratio = 0.2, min_indel_size = 35,
    maximum_local_variant_num = 100, skip_margin = 5000, 
    read_num_thres = 3, median_mapQ_thres = 20, max_overhang_size_thres = 100,
    max_control_read_num = 0, control_check_margin = 50, max_panel_read_num = 1, max_panel_sample_num = 1, debug = False):

    if debug: logger.setLevel(logging.DEBUG)

    sv_clusterer = Sv_clusterer(svtype, output_file, control_junction_bedpe = control_junction_bedpe, 
        control_panel_junction_bedpe = control_panel_junction_bedpe, bp_bed = bp_bed,
        cluster_margin_size = cluster_margin_size, size_margin_ratio = size_margin_ratio,
        maximum_local_variant_num = maximum_local_variant_num, skip_margin = skip_margin,
        read_num_thres = read_num_thres, median_mapQ_thres = median_mapQ_thres, 
        max_overhang_size_thres = max_overhang_size_thres, max_control_read_num = max_control_read_num,
        control_check_margin = control_check_margin, max_panel_read_num = max_panel_read_num, max_panel_sample_num = max_panel_sample_num)
    
    with gzip.open(input_file, 'rt') as hin: # , open(output_file, 'w') as hout:
        for line in hin:
            if sv_clusterer.svtype in ["insertion", "deletion"]: # insertion or deletion

                tchr1, tstart, tend, treadid, tsize, tdir, tinfo1 = line.rstrip('\n').split('\t')
                tstart1 = int(tstart) - indel_cluster_margin_size
                tend1 = int(tstart) + indel_cluster_margin_size
                tstart2 = int(tend) - indel_cluster_margin_size
                tend2 = int(tend) + indel_cluster_margin_size
                tchr2, tdir1, tdir2 = tchr1, '+', '-'
                tsize, tinfo2 = int(tsize), None

                if tsize < min_indel_size: continue
                
            elif sv_clusterer.svtype == "rearrangement": # rearrangement
                tchr1, tstart1, tend1, tchr2, tstart2, tend2, treadid, _, tdir1, tdir2, tinfo1, tinfo2 = \
                    line.rstrip('\n').split('\t')
                tstart1, tend1, tstart2, tend2 = int(tstart1), int(tend1), int(tstart2), int(tend2)
                tsize = None

            else:
                pass
                # logger.error("The svtype argument should be either of rearrangement, insertion or deletion.")
                # sys.exit(1)
 
            if sv_clusterer.temp_chr != tchr1:   
                sv_clusterer.next_pos_after_skip = 0
                sv_clusterer.temp_chr = tchr1
            if tend1 < sv_clusterer.next_pos_after_skip: continue

            # flush out existing cluster SVs wholse supporting reads have parsed alreadly 
            #   (considering the chromosome and coordinates).
            sv_clusterer.flush_sv_cluster_list(tchr1, tend1)

            # check the new SV can be merged into any existing SV clusters
            cret = sv_clusterer.check_mergeability(tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2, 
                treadid, tsize, tinfo1, tinfo2)           

            # create new SV cluster when new SV key cannot be merged into any existing SV clusteres
            if cret is None:
                sv_clusterer.add_new_cluster(tchr1, tstart1, tend1, tdir1, tchr2, tstart2, tend2, tdir2,
                    treadid, tsize, tinfo1, tinfo2)

            # check whether the number of locally clustered SVs and if it exceeds the threshould, 
            # then we ignore those SV clusteres and skip these regions.
            sv_clusterer.check_exceeding_local_variant_num(tchr1, tend1)

        sv_clusterer.flush_sv_cluster_list("EOF", 0)


if __name__ == "__main__":

    import sys
    input_file = sys.argv[1]
    output_file = sys.argv[2]
    svtype = sys.argv[3]
    control = sys.argv[4]
    bp = sys.argv[5]

    cluster_supporting_reads(input_file, output_file, svtype, control_junction_bedpe = control, bp_bed = bp, cluster_margin_size = 50)
             



#! /usr/bin/env python3

import sys, os, subprocess, shutil, statistics, logging
from collections import Counter
import pysam
import parasail
import resource

from .logger import get_logger

logger = get_logger(__name__)


class Consensus_generator(object):

    def __init__(self, output_file, use_racon, debug, max_memory_minimap2):

        self.tmp_dir = output_file + ".tmp_dir"
        os.makedirs(self.tmp_dir, exist_ok = True)

        self.hout = open(output_file, 'w')
        self.temp_key = None 
        self.readid2is_bp = {}
        self.temp_support_read_file_h = None
        self.use_racon = False if use_racon == False else True
        self.debug = debug
        self.parasail_error = []

        self.start_margin = 120
        self.min_inclusion_ratio = 0.9
        self.min_inclusion_count = 3
        self.max_memory_minimap2 = max_memory_minimap2

    def __del__(self):
        self.hout.close()
        if len(self.parasail_error) > 0:
            logger.debug(f"Alignment by parasail failed in {self.parasail_error}")

        if not self.debug:
            shutil.rmtree(self.tmp_dir)

    def initialize(self, temp_key):
        if self.temp_support_read_file_h is not None: self.temp_support_read_file_h.close()
        self.temp_key = temp_key
        self.temp_support_read_file_h = open(self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa", 'w')
        self.readid2is_bp = {}

    def add_support_read_seq(self, readid, seq, size):
        # skip if the seq corresponding to the input readid is already printed 
        if readid not in self.readid2is_bp:
            print(f">{readid}\n{seq}", file = self.temp_support_read_file_h)
            if size in ['-', '+']: # indication of break point type supporting read
                self.readid2is_bp[readid] = True
            else:
                self.readid2is_bp[readid] = False

    # function used for obtaining consensus sequence from multiple alignment result
    def get_consensus_from_mafft_result(self, input_file):

        id2seq = {}
        with open(input_file, 'r') as hin:
            for line in hin:
                line = line.rstrip('\n')
                if line.startswith('>'):
                    tid = line
                    id2seq[tid] = ''
                else:
                    id2seq[tid] = id2seq[tid] + line

        ind2bases = {}
        for tid in id2seq:
            seq = id2seq[tid]
            for i in range(len(seq)):
                if i not in ind2bases: ind2bases[i] = []
                ind2bases[i].append(seq[i])

        seq_len = len(list(ind2bases))
        consensus = ''
        for i in range(seq_len):
            mycounter = Counter(ind2bases[i] )
            consensus = consensus + mycounter.most_common()[0][0]

        consensus = consensus.replace('-', '').upper()

        return(consensus)


    def print_consensus_mafft(self):

        # if self.read_num >= 3: return

        hout_m = open(self.tmp_dir + '/' + self.temp_key + ".mafft_result.fa", 'w')
        subprocess.check_call(["mafft", self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa"], stdout = hout_m, stderr = subprocess.DEVNULL)
        hout_m.close()
                
        tconsensus = self.get_consensus_from_mafft_result(self.tmp_dir + '/' + self.temp_key + ".mafft_result.fa")
        print(self.temp_key + '\t' + tconsensus, file = self.hout)

        
    def generate_paf_file(self, query_fasta, target_fasta, output_file):

        user_matrix = parasail.matrix_create("ACGT", 2, -2)
        paf_rec_count = 0

        with open(target_fasta, 'r') as hin:
            for line in hin:
                if line.startswith('>'): 
                    tid = line.rstrip('\n').split(' ')[0].lstrip('>')
                else:
                    tseq = line.rstrip('\n')

        with open(query_fasta, 'r') as hin, open(output_file, 'w') as hout:
            for line in hin:
                if line.startswith('>'):
                    qid = line.rstrip('\n').lstrip('>')
                else:
                    qseq = line.rstrip('\n')
                    
                    res = parasail.ssw(qseq, tseq, 3, 1, user_matrix)
                    if res is not None:
                        print(f"{qid}\t{len(qseq)}\t{res.read_begin1}\t{res.read_end1}\t+\t" +
                            f"{tid}\t{len(tseq)}\t{res.ref_begin1}\t{res.ref_end1}\t*\t*\t60", file = hout)
                        paf_rec_count = paf_rec_count + 1
                    else:
                        self.parasail_error.append((qid, tid))

        return(paf_rec_count)


    def print_consensus_racon(self):

        target_flag = False
        with open(self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa", 'r') as hin, \
            open(self.tmp_dir + '/' + self.temp_key + ".tmp.seg.first.fa", 'w') as hout: 
            for line in hin:
                tid = line.rstrip('\n').lstrip('>')
                tseq = hin.readline().rstrip('\n')

                if not self.readid2is_bp[tid]:
                    print(f'>{tid}\n{tseq}', file = hout)
                    target_flag = True
                    break

        if target_flag == False:
            logger.debug(f"Template sequence could not be found for for {self.temp_key}")
            return

        paf_rec_count = self.generate_paf_file(self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
            self.tmp_dir + '/' + self.temp_key + ".tmp.seg.first.fa",
            self.tmp_dir + '/' + self.temp_key + ".parasail.paf")
                        
        if paf_rec_count < 3: 
            logger.debug(f"Not enough PAF records for the first round consensus generation for {self.temp_key}")
            return

        try: 
            with open(self.tmp_dir + '/' + self.temp_key + ".racon1.fa", 'w') as hout:
                subprocess.check_call(["racon", "-u", 
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
                    self.tmp_dir + '/' + self.temp_key + ".parasail.paf",
                    self.tmp_dir + '/' + self.temp_key + ".tmp.seg.first.fa"],
                    stdout = hout, stderr = subprocess.DEVNULL)
        except Exception as e:
            logger.warning(f'{e}')
            return

        with open(self.tmp_dir + '/' + self.temp_key + ".racon1.fa", 'r') as hin, \
            open(self.tmp_dir + "/" + self.temp_key + ".racon1.mod.fa", 'w') as hout:
            tid = hin.readline().lstrip('>').split(' ')[0]
            print(f">{tid}_2nd", file = hout)
            tseq = hin.readline().rstrip('\n')
            print(tseq, file = hout)

        paf_rec_count = self.generate_paf_file(self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
            self.tmp_dir + '/' + self.temp_key + ".racon1.mod.fa",
            self.tmp_dir + '/' + self.temp_key + ".parasail2.paf")
        
        if paf_rec_count < 3: 
            logger.debug(f"Not enough PAF records for the second round consensus generation for {self.temp_key}")
            return

        try:
            with open(self.tmp_dir + '/' + self.temp_key + ".racon2.fa", 'w') as hout:
                subprocess.check_call(["racon", "-u",
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
                    self.tmp_dir + '/' + self.temp_key + ".parasail2.paf",
                    self.tmp_dir + '/' + self.temp_key + ".racon1.mod.fa"],
                    stdout = hout, stderr = subprocess.DEVNULL)
        except Exception as e: 
            logger.warning(f'{e}')
            return

        with open(self.tmp_dir + "/" + self.temp_key + ".racon2.fa") as hin:
            header = hin.readline()
            tconsensus = hin.readline().rstrip('\n')
            print(f"{self.temp_key}\t{tconsensus}", file = self.hout)


    def print_consensus(self):

        if self.temp_support_read_file_h is not None: self.temp_support_read_file_h.close()
        if self.use_racon:
            self.print_consensus_racon()
        else:
            self.print_consensus_mafft()

    def preexec_fn(self):
        limit = self.max_memory_minimap2 * 1024 ** 3
        if "RLIMIT_AS" in resource.__dict__:
            resource.setrlimit(resource.RLIMIT_AS, (limit, limit))
        elif "RLIMIT_VMEM" in resource.__dict__:
            resource.setrlimit(resource.RLIMIT_VMEM, (limit, limit))

    def print_consensus_sbnd(self):

        if self.temp_support_read_file_h is not None: self.temp_support_read_file_h.close()

        try:
            with open(self.tmp_dir + '/' + self.temp_key + "_ava_minimap2.paf", 'w') as hout:
                subprocess.check_call(["minimap2", "-x", "ava-ont", 
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa"],
                    stderr = subprocess.DEVNULL, stdout = hout, preexec_fn = self.preexec_fn)
        except Exception as e:
            logger.warning(f'{e}')
            return


        readid2inclusion_count = {}
        with open(self.tmp_dir + '/' + self.temp_key + "_ava_minimap2.paf") as hin:
            for line in hin:
                row = line.rstrip('\n').split('\t')
                if int(row[2]) <= self.start_margin and (float(row[3]) - float(row[2])) / float(row[1]) >= self.min_inclusion_ratio:
                    if row[5] not in readid2inclusion_count: readid2inclusion_count[row[5]] = 0
                    readid2inclusion_count[row[5]] = readid2inclusion_count[row[5]] + 1
                if int(row[7]) <= self.start_margin and (float(row[8]) - float(row[7])) / float(row[6]) >= self.min_inclusion_ratio:
                    if row[0] not in readid2inclusion_count: readid2inclusion_count[row[0]] = 0
                    readid2inclusion_count[row[0]] = readid2inclusion_count[row[0]] + 1

        if len(readid2inclusion_count) == 0: return
        readid_max = max(readid2inclusion_count, key = readid2inclusion_count.get)
        if readid2inclusion_count[readid_max] < self.min_inclusion_count: return

        # first round racon
        with open(self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa", 'r') as hin, \
            open(self.tmp_dir + '/' + self.temp_key + "_ref.fa", 'w') as hout:
            seq_read_count = 0
            while True:
                tid = hin.readline().rstrip('\n').lstrip('>')
                tseq = hin.readline().rstrip('\n')
                if tid == readid_max:
                    print(f">{tid}\n{tseq}", file = hout)
                    break
                seq_read_count = seq_read_count + 1
                if tid == '' or seq_read_count > 10000: 
                    logger.warning(f"Something inconsistent happend when choosing template reads for {self.temp_key}")
                    return

        try:
            with open(self.tmp_dir + '/' + self.temp_key + "_ova_minimap2.paf", 'w') as hout:
                subprocess.check_call(["minimap2", "-x", "map-ont", self.tmp_dir + '/' + self.temp_key + "_ref.fa",
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa"], 
                    stderr = subprocess.DEVNULL, stdout = hout, preexec_fn = self.preexec_fn)
        except Exception as e:
            logger.warning(f'{e}')
            return

        paf_rec_count = 0
        with open(self.tmp_dir + '/' + self.temp_key + "_ova_minimap2.paf", 'r') as hin:
            for line in hin:
                paf_rec_count = paf_rec_count + 1

        if paf_rec_count < 3:
            logger.debug(f"Not enough PAF records for the first round consensus generation for {self.temp_key}")
            return

        consensus = ''
        try:
            with open(self.tmp_dir + '/' + self.temp_key + "_ref_polished.fa", 'w') as hout:
                subprocess.check_call(["racon", self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
                    self.tmp_dir + '/' + self.temp_key + "_ova_minimap2.paf",
                    self.tmp_dir + '/' + self.temp_key + "_ref.fa"], 
                    stderr = subprocess.DEVNULL, stdout = hout)
        except Exception as e:
            logger.warning(f'{e}')
            return

        with open(self.tmp_dir + '/' + self.temp_key + "_ref_polished.fa", 'r') as hin:
            for line in hin:
                if line.startswith('>'): continue
                consensus = consensus + line.rstrip('\n')

        if len(consensus) < 1000: return
                        
        # second round racon
        with open(self.tmp_dir + '/' + self.temp_key + "_ref_2nd.fa", 'w') as hout:
            print(f">{self.temp_key}_1st_polished_seq\n{consensus}", file = hout)

        try:
            with open(self.tmp_dir + '/' + self.temp_key + "_ova_minimap2_2nd.paf", 'w') as hout:
                subprocess.check_call(["minimap2", "-x", "map-ont", self.tmp_dir + '/' + self.temp_key + "_ref_2nd.fa",
                    self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa"],
                    stderr = subprocess.DEVNULL, stdout = hout, preexec_fn = self.preexec_fn)
        except Exception as e:
            logger.warning(f'{e}')
            return

        paf_rec_count = 0
        with open(self.tmp_dir + '/' + self.temp_key + "_ova_minimap2_2nd.paf", 'r') as hin:
            for line in hin:
                paf_rec_count = paf_rec_count + 1
            
        if paf_rec_count < 3:
            logger.debug(f"Not enough PAF records for the first round consensus generation for {self.temp_key}")
            return

        consensus = ''
        try:
            with open(self.tmp_dir + '/' + self.temp_key + "_ref_polished_2nd.fa", 'w') as hout:
                subprocess.check_call(["racon", self.tmp_dir + '/' + self.temp_key + ".supporting_read.fa",
                    self.tmp_dir + '/' + self.temp_key + "_ova_minimap2_2nd.paf",
                    self.tmp_dir + '/' + self.temp_key + "_ref_2nd.fa"],
                    stderr = subprocess.DEVNULL, stdout = hout)
        except Exception as e:
            logger.warning(f'{e}')
            return

        with open(self.tmp_dir + '/' + self.temp_key + "_ref_polished_2nd.fa", 'r') as hin:
            for line in hin:
                if line.startswith('>'): continue
                consensus = consensus + line.rstrip('\n')

        print(f"{self.temp_key}\t{consensus}", file = self.hout)



def generate_consensus(input_file, output_file, use_racon = False, debug = False, max_memory_minimap2 = 1): 

    if debug: logger.setLevel(logging.DEBUG)

    consensus_generator = Consensus_generator(output_file, use_racon, debug, max_memory_minimap2)
        
    with open(input_file, 'r') as hin:
        for line in hin:
            tkey, treadid, tsize, tseq = line.rstrip('\n').split('\t')
            # when new key appears transaction            
            if tkey != consensus_generator.temp_key:
                if consensus_generator.temp_key is not None:
                    consensus_generator.print_consensus()
                consensus_generator.initialize(tkey)

            consensus_generator.add_support_read_seq(treadid, tseq, tsize)

        if consensus_generator.temp_key is not None:
            consensus_generator.print_consensus()

    del consensus_generator 


def generate_consensus_sbnd(input_file, output_file, use_racon = False, debug = False, max_memory_minimap2 = 1):

    if debug: logger.setLevel(logging.DEBUG)

    # generate contig for single breakend
    consensus_generator_sbnd = Consensus_generator(output_file, use_racon, debug, max_memory_minimap2)
    with open(input_file, 'r') as hin:
        for line in hin:
            tkey, treadid, tsize, tseq = line.rstrip('\n').split('\t')
            # when new key appears transaction            
            if tkey != consensus_generator_sbnd.temp_key:
                if consensus_generator_sbnd.temp_key is not None:
                    consensus_generator_sbnd.print_consensus_sbnd()
                consensus_generator_sbnd.initialize(tkey)

            consensus_generator_sbnd.add_support_read_seq(treadid, tseq, tsize)

        if consensus_generator_sbnd.temp_key is not None:
            consensus_generator_sbnd.print_consensus_sbnd()

    del consensus_generator_sbnd

    



#! /usr/bin/env bash

import sys, csv
import pysam

input_file = sys.argv[1]
repeat_file = sys.argv[2]


repeat_tb = pysam.TabixFile(repeat_file)

with open(input_file, 'r') as hin:
    header = hin.readline().rstrip('\n')
    print(f'{header}\tIs_Simple_Repeat')
    for line in hin:
        F = line.rstrip('\n').split('\t')
        
        # check simplerepeat annotation
        repeat = "PASS"
        tabixErrorFlag = 0
        try:
            records = repeat_tb.fetch(F[0], int(F[1]) - 50, int(F[1]) + 50)
        except Exception as inst:
            print("%s: %s" % (type(inst), inst.args), file = sys.stderr)
            tabixErrorFlag = 1

        if tabixErrorFlag == 0:
            for record_line in records:
                record = record_line.split('\t')
                if int(F[1]) >= int(record[1]) - 50 and int(F[1]) <= int(record[2]) + 50:
                    repeat = record[0] + ':' + record[1] + '-' + record[2]

        print('\t'.join(F) + '\t' + repeat)





#! /usr/bin/env python

import sys, itertools

input_file = sys.argv[1]

consistent_pos_margin = 1.3
temp_rname = ""
temp_keys = []
temp_poses = []
temp_dirs = []
key2segment_size1 = {}
key2segment_size2 = {}

with open(input_file, 'r') as hin:
    for line in hin:
        F = line.rstrip('\n').split('\t')

        if F[1] != temp_rname:
            if temp_rname != "" and len(temp_keys) > 1:
                for i in range(0, len(temp_keys) - 1):

                    tchr1_1, tpos1_1, tdir1_1, tchr1_2, tpos1_2, tdir1_2, tinseq1 = temp_keys[i].split(',')
                    tchr2_1, tpos2_1, tdir2_1, tchr2_2, tpos2_2, tdir2_2, tinseq2 = temp_keys[i + 1].split(',')

                    jpos1, jpos2 = temp_poses[i], temp_poses[i + 1]
                    jdir1, jdir2 = temp_dirs[i], temp_dirs[i + 1]
            
                    if jdir1 == '+' and tinseq1 != "---": jpos1 = jpos1 + len(tinseq1)
                    if jdir2 == '+' and tinseq2 != "---": jpos2 = jpos2 - len(tinseq2)

                    if jdir1 == '+':
                        schr1_1, spos1_1, sdir1_1, schr1_2, spos1_2, sdir1_2 = \
                          tchr1_1, int(tpos1_1), tdir1_1, tchr1_2, int(tpos1_2), tdir1_2
                    else:
                        schr1_2, spos1_2, sdir1_2, schr1_1, spos1_1, sdir1_1 = \
                          tchr1_1, int(tpos1_1), tdir1_1, tchr1_2, int(tpos1_2), tdir1_2

                    if jdir2 == '+':
                        schr2_1, spos2_1, sdir2_1, schr2_2, spos2_2, sdir2_2 = \
                          tchr2_1, int(tpos2_1), tdir2_1, tchr2_2, int(tpos2_2), tdir2_2 
                    else:
                        schr2_2, spos2_2, sdir2_2, schr2_1, spos2_1, sdir2_1 = \
                          tchr2_1, int(tpos2_1), tdir2_1, tchr2_2, int(tpos2_2), tdir2_2

                    sinseq1, sinseq2 = tinseq1, tinseq2

                    # check positional consistency
                    # if spos1_2 == 25829678 or spos2_1 == 25829678:
                    #     import pdb; pdb.set_trace()

                    segment_size = '---'
                    consistent_flag = False
                    segment_dir = '---'
                    if schr1_2 == schr2_1:
                        if sdir1_2 == '-' and sdir2_1 == '+': 
                            if spos2_1 - spos1_2 <= consistent_pos_margin * (jpos2 - jpos1) and \
                              spos2_1 - spos1_2 >= (1 / consistent_pos_margin) * (jpos2 - jpos1):
                                consistent_flag = True
                                segment_size = spos2_1 - spos1_2
                                segment_dir = '+'
                        elif sdir1_2 == '+' and sdir2_1 == '-':
                            if spos1_2 - spos2_1 <= consistent_pos_margin * (jpos2 - jpos1) and \
                              spos1_2 - spos2_1 >= (1 / consistent_pos_margin) * (jpos2 - jpos1):
                                consistent_flag = True
                                segment_dir = '-'
                                segment_size = spos1_2 - spos2_1

                    if consistent_flag == True and segment_dir == '-':
                        schr1_1, schr2_2 = schr2_2, schr1_1
                        spos1_1, spos2_2 = spos2_2, spos1_1
                        sdir1_1, sdir2_2 = sdir2_2, sdir1_1
                        schr1_2, schr2_1 = schr2_1, schr1_2
                        spos1_2, spos2_1 = spos2_1, spos1_2
                        sdir1_2, sdir2_1 = sdir2_1, sdir1_2
                        sinseq1, sinseq2 = sinseq2, sinseq1
    
                    if consistent_flag == False:
                        if schr1_2 < schr2_1 or (schr1_2 == schr2_1 and spos1_2 > spos2_1):
                            schr1_1, schr2_2 = schr2_2, schr1_1
                            spos1_1, spos2_2 = spos2_2, spos1_1
                            sdir1_1, sdir2_2 = sdir2_2, sdir1_1
                            schr1_2, schr2_1 = schr2_1, schr1_2
                            spos1_2, spos2_1 = spos2_1, spos1_2
                            sdir1_2, sdir2_1 = sdir2_1, sdir1_2
                            sinseq1, sinseq2 = sinseq2, sinseq1

                    key = '\t'.join([schr1_1, str(spos1_1), sdir1_1, sinseq1, schr1_2, str(spos1_2), sdir1_2,
                                     schr2_1, str(spos2_1), sdir2_1, sinseq2, schr2_2, str(spos2_2), sdir2_2, str(consistent_flag)])

                    if key not in key2segment_size1:
                        key2segment_size1[key], key2segment_size2[key] = str(segment_size), []

                    # key2segment_size1[key].append(str(segment_size))
                    key2segment_size2[key].append(str(jpos2 - jpos1))

                    """
                    print('\t'.join([schr1_1, str(spos1_1), sdir1_1, sinseq1, schr1_2, str(spos1_2), sdir1_2,
                                     schr2_1, str(spos2_1), sdir2_1, sinseq2, schr2_2, str(spos2_2), sdir2_2,
                                     str(consistent_flag), str(segment_size), str(jpos2 - jpos1)]))
                    """

            temp_rname = F[1]
            temp_keys = []
            temp_poses = []
            temp_dirs = []

        temp_keys.append(F[0])
        temp_poses.append(int(F[2]))
        temp_dirs.append(F[3])


for key in key2segment_size1:
    segment_size1 = key2segment_size1[key]
    segment_size2 = ','.join(key2segment_size2[key])

    print('\t'.join([key, segment_size1, segment_size2]))




#! /usr/bin/env python

import sys
from .arg_parser import create_parser

def main():

    parser = create_parser()
    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args.func(args)




#! /usr/bin/env python3

import argparse

from .run import *
from .version import __version__

def create_parser():

    parser = argparse.ArgumentParser(prog = "nanomonsv")

    parser.add_argument("--version", action = "version", version = "%(prog)s " + __version__)

    subparsers = parser.add_subparsers()

    ##########
    # parse
    parse = subparsers.add_parser("parse",
                                help = "Parse supporting reads for candidate structural variations")


    parse.add_argument("alignment_file", default = None, type = str,
                     help = "Path to input BAM or CRAM file")

    parse.add_argument("output_prefix", type = str,
                       help = "Prefix of output files")

    parse.add_argument("--reference_fasta", metavar = "reference.fa", default = None, type = str,
                     help = "the path to the reference genome sequence")

    parse.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    parse.add_argument("--split_alignment_check_margin", default = 50, type = int,
                       help = "Two split alignments whose margin sizes are no more than this value is counted as candidate breakpoint (default: 50)")

    parse.add_argument("--minimum_breakpoint_ambiguity", default = 20, type = int,
                       help = "Sizes of ambiguities of breakpoint positions from the observed ones (default: 20)") 

    parse.set_defaults(func = parse_main)
    ##########
   
    ##########
    # merge_control
    merge_control = subparsers.add_parser("merge_control", 
                                          help = "Merge control panel files from parse results")

    merge_control.add_argument("prefix_list_file", default = None, type = str,
                               help = "Path to prefix for the parse function for control panels")
        
    merge_control.add_argument("output_prefix", default = None, type = str,
                               help = "Prefix of merged control files")

    merge_control.set_defaults(func = merge_control_main)
    ##########

    ##########
    # get
    get = subparsers.add_parser("get",
                                help = "List up reliable structural variations with refined breakpoint positions")

    get.add_argument("tumor_prefix", type = str,
                      help = "Prefix of tumor data processed in parse step")
       
    get.add_argument("tumor_bam", default = None, type = str,
                      help = "Path to tumor alignment (BAM or CRAM) file")
 
    get.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                     help = "Path to the reference genome sequence")

    get.add_argument("--control_prefix", type = str, # required = True,
                     help = "Prefix of matched control data processed in parse step")

    get.add_argument("--control_bam", type = str, # required = True,
                     help = "Path to control alignment (BAM or CRAM) file")

    get.add_argument("--control_panel_prefix", type = str, 
                     help = "Prefix of non-matched control panel data processed in merge_control step")

    get.add_argument("--simple_repeat_bed", type = str, default = None,
                     help = "Path to simple repeat bed file")

    get.add_argument("--min_tumor_variant_read_num", default = 3, type = int,
                     help = "Minimum required supporting read number for a tumor sample (default: 3)")

    get.add_argument("--min_tumor_VAF", default = 0.05, type = float,
                     help = "Minimum required variant allele frequency for a tumor sample (default: 0.05)")

    get.add_argument("--max_control_variant_read_num", default = 1, type = int,
                     help = "Maximum allowed supporting read number for a control sample (default: 1)")
        
    get.add_argument("--max_control_VAF", default = 0.03, type = float,
                     help = "Maximum allowed variant allele frequeycy for a control sample (default: 0.03)")

    get.add_argument("--min_indel_size", default = 50, type = int,
                     help = "Minimum indel size for the output (default: 50)")

    get.add_argument("--max_panel_read_num", default = 1, type = int,
                     help= "Maximum allowed supporting read number for a nonmatched control sample (default: 1)")

    get.add_argument("--max_panel_sample_num", default = 0, type = int,
                     help= "Maximum allowed sample number for a nonmatched control sample (default: 0)")

    get.add_argument("--cluster_margin_size", default = 100, type = int,
                     help = "Two breakpoints are margined if they are within this threshould value (default: 100)")

    get.add_argument("--median_mapQ_thres", default = 20, type = int,
                     help = "Threshould for median mapping quality (default: 20)")

    get.add_argument("--max_overhang_size_thres", default = 100, type = int,
                     help = "Threshould for maximum overhang size (default: 100)")

    get.add_argument("--check_read_max_num", default = 500, type = int,
                     help = "The maximum number of reads to check per breakpoint in the phase of realignment validation (default: 500)")

    get.add_argument("--var_read_min_mapq", default = 0, type = int,
                     help = "Threshould for mapping quality in validate step (default: 0)")

    get.add_argument("--validation_score_ratio_thres", default = 1.2, type = float,
                     help = "Threshould for threshould for SV segment validation by alignment")

    get.add_argument("--sw_jump_params", nargs = 4, default = [1, 3, 3, 2], type = int,
                     help = "Parameters (match score, mismatch penalty, gap penalty, insertion penalty) for one-time smith-waterman algorithm (default: [1, 3, 3, 2]")

    # get.add_argument("--use_ssw_lib", default = False, action = 'store_true',
    #                  help = "Use SSW Library. This is for backward comaptibility, and may be removed in the future (default: False)")

    get.add_argument("--qv10", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 10. Recommended for ONT data called by Guppy before version 5")

    get.add_argument("--qv15", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 15. Recommended for ONT data called by Guppy version 5, 6.")

    get.add_argument("--qv20", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 20. Recommended for ONT data with Q20+ chemistry.")

    get.add_argument("--qv25", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality above 25. Recommended for PacBio Hifi data.")

    get.add_argument("--use_racon", default = False, action = 'store_true',
                     help = "Use racon for error correction of clustered putative supporting reads (default: False)")

    get.add_argument("--single_bnd", default = False, action = 'store_true',
                     help = "Generate single end breakpoints (default: False)")

    # get.add_argument("--control_read_num_thres", default = 0, type = int,
    #                  help = "Filter if the number of supporting reads for the control sample is larger than this value")

    # get.add_argument("--threads", default = 1, type = int,
    #                  help = "Number of parallel threads to use (not recommended) (default: 1)")

    get.add_argument("--processes", default = 1, type = int,
                     help = "Number of parallel processes to use (default: 1)")

    get.add_argument("--sort_option", type = str, default = "-S 1G", 
                     help = "Options for Linux sort command (default: '-S 1G')")

    get.add_argument("--max_memory_minimap2", type = int, default = 2, 
                     help = "Maximum memory size (Gbyte) for minimap2 (default: 2)")

    get.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files (default: False)")

    get.set_defaults(func = get_main)
    ##########

    ##########
    # validate 
    validate = subparsers.add_parser("validate",
                                     help = "Validate GenomonSV format SV list using nanopore sequence reads")
 
    validate.add_argument("sv_list_file", default = None, type = str,
                          help = "Path to GenomonSV format SV list file")

    validate.add_argument("tumor_bam", default = None, type = str,
                          help = "Path to tumor BAM file")

    validate.add_argument("output", type = str,
                       help = "Path to output file")

    validate.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                          help = "Path to the reference genome sequence")

    validate.add_argument("--control_bam", default = None, type = str,
                          help = "Path to control BAM file")

    validate.add_argument("--var_read_min_mapq", default = 40, type = int,
                          help = "Threshould for mapping quality in validate step")

    validate.add_argument("--validation_score_ratio_thres", default = 1.2, type = float, 
                     help = "Threshould for threshould for SV segment validation by alignment")

    validate.add_argument("--qv10", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 10. Recommended for ONT data called by Guppy before version 5")
    
    validate.add_argument("--qv15", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 15. Recommended for ONT data called by Guppy version 5, 6.")
    
    validate.add_argument("--qv20", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality of around 20. Recommended for ONT data with Q20+ chemistry.")

    validate.add_argument("--qv25", default = False, action = 'store_true',
                     help = "Parameter preset for sequencing data with a base quality above 25. Recommended for PacBio Hifi data.")

    validate.add_argument("--sort_option", metavar = "-S 1G", type = str, default = "-S 2G", 
                     help = "options for sort command")

    validate.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    validate.set_defaults(func = validate_main)
    ##########

    ##########
    # insert_classify
    insert_classify = subparsers.add_parser("insert_classify",
                                            help = "Classify long insertion into LINE1, Alu, SVA, and so on")
    
    insert_classify.add_argument("sv_list_file", default = None, type =str,
                                 help = "Path to nanomonsv get result file")
    
    insert_classify.add_argument("output_file", type = str,
                                 help = "Path to output file")

    insert_classify.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                                 help = "Path to the reference genome sequence")

    # insert_classify.add_argument("--grc", default = False, action = 'store_true',
    #                              help = "Deprecated. This is not used any more. Convert chromosome names to Genome Reference Consortium nomenclature (default: %(default)s)")
    insert_classify.add_argument("gtf_file", metavar = "gencode.gtf.gz", default = None, type = str,
                                 help = "Path to GFT file for transcript")

    # insert_classify.add_argument("--genome_id", choices = ["hg19", "hg38", "mm10"], default = "hg38",
    #                              help = "Genome id used for selecting UCSC-GRC chromosome name corresponding files (default: %(default)s)")

    insert_classify.add_argument("LINE1_db", default = None, type = str,
                                 help = "Path to LINE1 position file")

    insert_classify.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    insert_classify.set_defaults(func = insert_classify_main)
    ##########
    
    ##########
    # connect
    connect = subparsers.add_parser("connect",
                                    help = "Connect called SVs by support reads")

    connect.add_argument("nanomonsv_result_file", default = None, type = str,
                         help = "Path to nanomonsv get result file")

    connect.add_argument("support_read_file", type = str,
                         help = "PATH to support read file")
    
    connect.add_argument("output_prefix", type = str,
                         help = "Prefix of output files")

    connect.add_argument("--consistent_pos_margin", default = 1.3, type = float,
                         help = "Consistency coefficient of distances of read and reference between two SVs")
    
    connect.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    connect.set_defaults(func = connect_main)
    ##########

    return parser


    



#! /usr/bin/env python3

import logging

def get_logger(mod_name):

    logger = logging.getLogger(mod_name)
    logger.propagate = False

    if not logger.hasHandlers(): 
        ch = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    logger.setLevel(logging.INFO)

    return logger




#! /usr/bin/env python

import sys

input_file = sys.argv[1]

with open(input_file, 'r') as hin:
    for line in hin:
        F = line.rstrip('\n').split('\t')
        label_info = F[3].split(',')
        label_info[0], label_info[1], label_info[2] = F[0], F[1], F[2]
        F[3] = ','.join(label_info)
        print('\t'.join(F))
        



#! /usr/bin/env python3

import sys, csv
import pysam


def filter_indel_in_simple_repeat(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, simple_repeat_tb, simple_repeat_dist_margin = 30):

    if tchr1 == tchr2 and tdir1 == '+' and tdir2 == '-':
        sv_size = tpos2 - tpos1 + len(tinseq) - 1

        tabix_error_flag = False
        try:
            records = simple_repeat_tb.fetch(tchr1, max(tpos1 - simple_repeat_dist_margin + 1, 0), 
                tpos1 + simple_repeat_dist_margin)
        except Exception as inst:
            print(f'{type(inst)}: {inst.args}', file = sys.stderr)
            tabix_error_flag = True

        if tabix_error_flag == False:
            for record_line in records:
                record = record_line.split('\t')
                if tpos1 >= int(record[1]) - simple_repeat_dist_margin and \
                    int(tpos2) <= int(record[2]) + simple_repeat_dist_margin:
                    return True

        return False


def post_filter_main(args):

    simple_repeat_tb = pysam.TabixFile(args.simple_repeat_bed) if args.simple_repeat_bed is not None else None
    
    with open(args.sv_list_file, 'r') as hin, open(args.output_file, 'w') as hout:
        dreader = csv.DictReader(hin, delimiter = '\t')
        header = dreader.fieldnames
        print('\t'.join(header), file = hout)

        for F in dreader:
            tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq = F["Chr_1"], int(F["Pos_1"]), F["Dir_1"], F["Chr_2"], int(F["Pos_2"]), F["Dir_2"], F["Inserted_Seq"]

            if args.simple_repeat_bed is not None:
                simple_repeat_flag = filter_indel_in_simple_repeat(tchr1, tpos1, tdir1, tchr2, tpos2, tdir2, tinseq, simple_repeat_tb)
            else:
                simple_repeat_flag = False

            if simple_repeat_flag:
                if F["Is_Filter"] == "PASS": 
                    F["Is_Filter"] = "Simple_repeat"
                else:
                    F["Is_Filter"] = F["Is_Filter"] + ';' + "Simple_repeat"

            print('\t'.join(F.values()), file = hout)
 

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(prog = "nanomonsv_simple_repeat_annot",
        description = "Add simple repeat annotation to the result of nanomonsv")

    parser.add_argument("sv_list_file", type = str,
                        help = "Path to the nanomonsv result file")

    parser.add_argument("output_file", type = str,
                        help = "Path to the output file")

    parser.add_argument("simple_repeat_bed", metavar = "simpleRepeat.bed.gz", type = str, default = None,
                        help = "Path to the tabix indexed simple repeat bed file")

    args = parser.parse_args()

    post_filter_main(args)


